"""
SCTE35 Splice Commands
"""

from .bitn import BitBin
from .base import SCTE35Base
from .xml import Node


class SpliceCommand(SCTE35Base):
    """
    Base class, not used directly.
    """

    def __init__(self, bites=None):
        self.command_length = 0
        self.command_type = None
        self.name = None
        self.bites = bites

    def decode(self):
        """
        default decode method
        """

    def _set_len(self, start, end):
        """
        _set_len sets
        self.command_length
        """
        self.command_length = (start - end) >> 3

    def encode(self, nbin=None):
        """
        encode
        """
        nbin = self._chk_nbin(nbin)
        return nbin.bites

    def from_xml(self, gonzo):
        """
        load a SpliceCommand from xml - default method does nothing
        """


class BandwidthReservation(SpliceCommand):
    """
    Table 12 - bandwidth_reservation()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.command_type = 7
        self.name = "Bandwidth Reservation"

    def decode(self):
        """
        BandwidthReservation.decode method
        """

    def xml(self, ns="scte35"):
        """
        create XML Node of type BandwidthReservation
        """
        return Node("BandwidthReservation", ns=ns)


class PrivateCommand(SpliceCommand):
    """
    Table 13 - private_command
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.command_type = 255
        self.name = "Private Command"
        self.identifier = None
        self.private_bytes = None

    def decode(self):
        """
        PrivateCommand.decode method
        """
        self.identifier = int.from_bytes(
            self.bites[:4], byteorder="big"
        )  # 4 bytes = 32 bits
        self.private_bytes = self.bites[4:]

    def encode(self, nbin=None):
        """
        encode private command
        """
        nbin = self._chk_nbin(nbin)
        self._chk_var(int, nbin.add_int, "identifier", 32)  # 4 bytes = 32 bits
        nbin.add_bites(self.private_bytes)
        return nbin.bites

    def xml(self, ns="scte35"):
        """
        create XML Node of type PrivateCommand
        """
        attrs = {"identifier": self.identifier}
        pc = Node("PrivateCommand", attrs=attrs)
        pc.add_child(Node("PrivateBytes", value=self.private_bytes.hex(), ns=ns))
        return pc


##    def from_xml(self, gonzo):
##        """
##        load a PrivateCommand from XML
##        """
##        self.identifier = gonzo["PrivateCommand"]["identifier"]
##        if "PrivateBytes" in gonzo and "private_bytes" in gonzo["PrivateBytes"]:
##            self.private_bytes = bytes.fromhex((gonzo["PrivateBytes"]["private_bytes"]))
##


class SpliceNull(SpliceCommand):
    """
    Table 8 - splice_null()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.command_type = 0
        self.name = "Splice Null"

    def xml(self, ns="scte35"):
        return Node("SpliceNull", ns=ns)


class TimeSignal(SpliceCommand):
    """
    Table 11 - time_signal()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.command_type = 6
        self.name = "Time Signal"
        self.time_specified_flag = None
        self.pts_time = None

    def decode(self):
        """
        TimeSignal.decode method
        """
        bitbin = BitBin(self.bites)
        start = bitbin.idx
        self._splice_time(bitbin)
        self._set_len(start, bitbin.idx)

    def encode(self, nbin=None):
        """
        encode converts TimeSignal vars
        to bytes
        """
        nbin = self._chk_nbin(nbin)
        self._encode_splice_time(nbin)
        return nbin.bites

    def _splice_time(self, bitbin):
        """
        _splice_time Table 14 - splice_time()
        """
        self.time_specified_flag = bitbin.as_flag(1)
        if self.time_specified_flag:
            bitbin.forward(6)
            pts_time_ticks = bitbin.as_int(33)
            self.pts_time = self.as_90k(pts_time_ticks)
        else:
            bitbin.forward(7)

    def _encode_splice_time(self, nbin):
        """
        _encode_splice_time Table 14 - splice_time()
        """
        self._chk_var(bool, nbin.add_flag, "time_specified_flag", 1)
        if self.time_specified_flag:
            nbin.reserve(6)
            if not isinstance(self.pts_time, float):
                raise ValueError(
                    "\033[7mA float for pts_time for the splice command is required.\033[27m"
                )
            self.pts_time = round(self.pts_time, 6)
            nbin.add_int(int(self.as_ticks(self.pts_time)), 33)
        else:
            nbin.reserve(7)

    def xml(self, ns="scte35"):
        """
        xml return TimeSignal as an xml node
        """
        ts = Node("TimeSignal", ns=ns)
        if self.has("pts_time"):
            if self.pts_time:
                self.pts_time = round(self.pts_time, 6)
                st = Node(
                    "SpliceTime",
                    attrs={"pts_time": self.as_ticks(self.pts_time)},
                    ns=ns,
                )
                ts.add_child(st)
        return ts


class SpliceInsert(TimeSignal):
    """
    Table 10 - splice_insert()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.command_type = 5
        self.name = "Splice Insert"
        self.break_auto_return = None
        self.break_duration = None
        self.splice_event_id = None
        self.splice_event_cancel_indicator = None
        self.out_of_network_indicator = None
        self.program_splice_flag = None
        self.duration_flag = None
        self.splice_immediate_flag = None
        self.event_id_compliance_flag = None
        self.unique_program_id = None
        self.avail_num = None
        self.avails_expected = None

    def decode_break_duration(self, bitbin):
        """
        break_duration Table 15 - break_duration()
        """
        self.break_auto_return = bitbin.as_flag(1)
        bitbin.forward(6)
        break_duration_ticks = bitbin.as_int(33)
        self.break_duration = self.as_90k(break_duration_ticks)

    def decode(self):
        """
        decode SpliceInsert
        """
        bitbin = BitBin(self.bites)
        start = bitbin.idx
        self.splice_event_id = bitbin.as_int(32)
        self.splice_event_cancel_indicator = bitbin.as_flag(1)
        bitbin.forward(7)
        if not self.splice_event_cancel_indicator:
            self.out_of_network_indicator = bitbin.as_flag(1)
            self.program_splice_flag = bitbin.as_flag(1)
            self.duration_flag = bitbin.as_flag(1)
            self.splice_immediate_flag = bitbin.as_flag(1)
            self.event_id_compliance_flag = bitbin.as_flag(1)
            bitbin.forward(3)
            if self.program_splice_flag:
                if not self.splice_immediate_flag:
                    self._splice_time(bitbin)
            if self.duration_flag:
                self.decode_break_duration(bitbin)
            self.unique_program_id = bitbin.as_int(16)
            self.avail_num = bitbin.as_int(8)
            self.avails_expected = bitbin.as_int(8)
        self._set_len(start, bitbin.idx)

    def encode(self, nbin=None):
        """
        SpliceInsert.encode
        """
        nbin = self._chk_nbin(nbin)
        self._chk_var(int, nbin.add_int, "splice_event_id", 32)
        self._chk_var(bool, nbin.add_flag, "splice_event_cancel_indicator", 1)
        nbin.forward(7)
        if not self.splice_event_cancel_indicator:
            self._chk_var(bool, nbin.add_flag, "out_of_network_indicator", 1)
            self._chk_var(bool, nbin.add_flag, "program_splice_flag", 1)
            self._chk_var(bool, nbin.add_flag, "duration_flag", 1)
            self._chk_var(bool, nbin.add_flag, "splice_immediate_flag", 1)
            self._chk_var(bool, nbin.add_flag, "event_id_compliance_flag", 1)
            nbin.forward(3)
            if self.program_splice_flag:
                if not self.splice_immediate_flag:
                    self._encode_splice_time(nbin)
            if self.duration_flag:
                self.encode_break_duration(nbin)
            self._chk_var(int, nbin.add_int, "unique_program_id", 16)
            self._chk_var(int, nbin.add_int, "avail_num", 8)
            self._chk_var(int, nbin.add_int, "avails_expected", 8)
        return nbin.bites

    def encode_break_duration(self, nbin):
        """
        SpliceInsert._encode_break(nbin) is called
        if SpliceInsert.duration_flag is set
        """
        self._chk_var(bool, nbin.add_flag, "break_auto_return", 1)
        nbin.forward(6)
        nbin.add_int(self.as_ticks(self.break_duration), 33)

    def xml(self, ns="scte35"):
        """
        xml return the SpliceInsert instance as an xml node.
        """
        si_attrs = {
            "splice_event_id": self.splice_event_id,
            "splice_event_cancel_Indicator": self.splice_event_cancel_indicator,
            "splice_immediate_flag": self.splice_immediate_flag,
            "event_id_compliance_flag": self.event_id_compliance_flag,
            "avail_num": self.avail_num,
            "avails_expected": self.avails_expected,
            "out_of_network_indicator": self.out_of_network_indicator,
            "unique_program_id": self.unique_program_id,
        }
        for k, v in si_attrs.items():
            if v is None:
                raise ValueError(f"\033[7mSpliceInsert.{k} needs to be set\033[27m")
        si = Node("SpliceInsert", attrs=si_attrs, ns=ns)
        if self.pts_time:
            prgm = Node("Program", ns=ns)
            st = Node(
                "SpliceTime", attrs={"ptsTime": self.as_ticks(self.pts_time)}, ns=ns
            )
            prgm.add_child(st)
            si.add_child(prgm)
        if self.break_duration:
            bd_attrs = {
                "auto_return": self.break_auto_return,
                "duration": self.as_ticks(self.break_duration),
            }
            bd = Node("BreakDuration", attrs=bd_attrs, ns=ns)
            si.add_child(bd)
        return si


# table 7
command_map = {
    0: SpliceNull,
    5: SpliceInsert,
    6: TimeSignal,
    7: BandwidthReservation,
    255: PrivateCommand,
}



"""
fu.Cue Class
"""

from base64 import b64decode, b64encode
import json
from .spare import print2
from .bitn import NBin
from .base import SCTE35Base
from .section import SpliceInfoSection
from .commands import command_map
from .descriptors import splice_descriptor, descriptor_map
from .crc import crc32
from .xml import Node
from .segmentation import table22
from .sxp import SuperXmlParser


class Cue(SCTE35Base):
    """
    The fu.Splice class handles parsing
    SCTE 35 message strings.
    Example usage:

    >>>> import fu
    >>>> Base64 = "/DAvAAAAAAAA///wBQb+dGKQoAAZAhdDVUVJSAAAjn+fCAgAAAAALKChijUCAKnMZ1g="
    >>>> cue = fu.Cue(Base64)

    # cue.decode() returns True on success,or False if decoding failed

    >>>> cue.decode()
    True

    # After Calling cue.decode() the instance variables can be accessed via dot notation.

    >>>> cue.command
    {'command_length': 5, 'name': 'Time Signal', 'time_specified_flag': True,
    'pts_time': 21695.740089}

    >>>> cue.command.pts_time
    21695.740089

    >>>> cue.info_section.table_id

    '0xfc'

    """

    def __init__(self, data=None, packet_data=None):
        """
        data may be packet bites or encoded string
        packet_data is a instance passed from a Stream instance
        """
        self.command = None
        self.descriptors = []
        self.info_section = SpliceInfoSection()
        self.bites = None
        if data:
            self.bites = self._mk_bits(data)
        self.packet_data = packet_data
        self.dash_data = None

    def __repr__(self):
        return str(self.__dict__)

    # decode

    def decode(self):
        """
        Cue.decode() parses for SCTE35 data
        """
        bites = self.bites
        self.descriptors = []
        while bites:
            bites = self.mk_info_section(bites)
            bites = self._set_splice_command(bites)
            bites = self._mk_descriptors(bites)
            crc = hex(int.from_bytes(bites[0:4], byteorder="big"))
            self.info_section.crc = crc
            return True
        return False

    def _descriptor_loop(self, loop_bites):
        """
        Cue._descriptor_loop parses all splice descriptors
        """
        tag_n_len = 2
        while len(loop_bites) > tag_n_len:
            spliced = splice_descriptor(loop_bites)
            if not spliced:
                return
            sd_size = tag_n_len + spliced.descriptor_length
            loop_bites = loop_bites[sd_size:]
            del spliced.bites
            self.descriptors.append(spliced)

    def _get_dash_data(self, scte35_dict):
        if self.dash_data:
            scte35_dict["dash_data"] = self.dash_data
        return scte35_dict

    def _get_packet_data(self, scte35_dict):
        if self.packet_data:
            scte35_dict["packet_data"] = self.packet_data.get()
        return scte35_dict

    def get(self):
        """
        Cue.get returns the SCTE-35 Cue
        data as a dict of dicts.
        """
        if self.command and self.info_section:
            scte35 = {
                "info_section": self.info_section.get(),
                "command": self.command.get(),
                "descriptors": self.get_descriptors(),
            }
            scte35 = self._get_dash_data(scte35)
            scte35 = self._get_packet_data(scte35)
            return scte35
        return False

    def get_descriptors(self):
        """
        Cue.get_descriptors returns a list of
        SCTE 35 splice descriptors as dicts.
        """
        return [d.get() for d in self.descriptors]

    def bytes(self):
        """
        get_bytes returns Cue.bites
        """
        return self.bites

    @staticmethod
    def fix_bad_b64(data):
        """
        fix_bad_b64 fixes bad padding on Base64
        """
        while len(data) % 4 != 0:
            data = data + "="
        return data

    def _int_bits(self, data):
        """
        _int_bits convert a SCTE-35 Cue from integer to bytes.
        """
        length = data.bit_length() >> 3
        bites = int.to_bytes(data, length, byteorder="big")
        return bites

    def _hex_bits(self, data):
        """
        _hex_bits convert a SCTE-35 Cue from hex to bytes.
        """
        try:
            i = int(data, 16)
            i_len = i.bit_length() >> 3
            bites = int.to_bytes(i, i_len, byteorder="big")
            return bites
        except (LookupError, TypeError, ValueError):
            if data[:2].lower() == "0x":
                data = data[2:]
            if data[:2].lower() == "fc":
                return bytes.fromhex(data)
        return b""

    def _b64_bits(self, data):
        """
        _b64_bits decode base64 to bytes
        """
        try:
            return b64decode(self.fix_bad_b64(data))
        except (LookupError, TypeError, ValueError):
            return data

    def _str_bits(self, data):
        try:
            self.load(data)
            return self._mk_load(data)
        except:
            hex_bits = self._hex_bits(data)
            if hex_bits:
                return hex_bits
        return self._b64_bits(data)

    def _mk_load(self, data):
        """
         _mk_load sets self.bytes when loading
        data. Encode is called to set missing fields
        when possible and re-calc the length vars and crc.
        """
        # if self.load(data):
        # bites = self.bites
        # self.encode()
        # return bites
        # return seldata
        self.load(data)

    def _mk_bits(self, data):
        """
        cue._mk_bits Converts
        Hex and Base64 strings into bytes.
        """
        if isinstance(data, bytes):
            return self.idxsplit(data, b"\xfc")
        if isinstance(data, int):
            return self._int_bits(data)
        if isinstance(data, dict):
            return self.load(data)
        if isinstance(data, Node):
            return self.load(data.mk())
        if isinstance(data, str):
            ##            if data.isdigit():
            ##                return self._int_bits(int(data))
            return self._str_bits(data)

    def _mk_descriptors(self, bites):
        """
        Cue._mk_descriptors parses
        Cue.info_section.descriptor_loop_length,
        then call Cue._descriptor_loop
        """
        if len(bites) < 2:
            return False
        dll = (bites[0] << 8) | bites[1]
        self.info_section.descriptor_loop_length = dll
        bites = bites[2:]
        self._descriptor_loop(bites[:dll])
        return bites[dll:]

    def mk_info_section(self, bites):
        """
        Cue.mk_info_section parses the
        Splice Info Section
        of a SCTE35 cue.
        """
        info_size = 14
        info_bites = bites[:info_size]
        self.info_section.decode(info_bites)
        return bites[info_size:]

    def _set_splice_command(self, bites):
        """
        Cue._set_splice_command parses
        the command section of a SCTE35 cue.
        """
        sct = self.info_section.splice_command_type
        if sct not in command_map:
            return False
        iscl = self.info_section.splice_command_length
        cmd_bites = bites[:iscl]
        self.command = command_map[sct](cmd_bites)
        self.command.command_length = iscl

        self.command.decode()
        del self.command.bites
        return bites[iscl:]

    # encode related

    def _assemble(self):
        dscptr_bites = self._unloop_descriptors()
        dll = len(dscptr_bites)
        self.info_section.descriptor_loop_length = dll
        cmd_bites = self.command.encode()
        cmdl = self.command.command_length = len(cmd_bites)
        self.info_section.splice_command_length = cmdl
        self.info_section.splice_command_type = self.command.command_type
        # 11 bytes for info section + command + 2 descriptor loop length
        # + descriptor loop + 4 for crc
        self.info_section.section_length = 11 + cmdl + 2 + dll + 4
        self.bites = self.info_section.encode()
        self.bites += cmd_bites
        self.bites += int.to_bytes(
            self.info_section.descriptor_loop_length, 2, byteorder="big"
        )
        self.bites += dscptr_bites

    def base64(self):
        """
        base64 Cue.base64() converts SCTE35 data
        to a base64 encoded string.
        """
        if self.command:
            self._assemble()
            self._encode_crc()
            return b64encode(self.bites).decode()
        return False

    def encode(self):
        """
        encode is an alias for base64
        """
        return self.base64()

    def int(self):
        """
        int returns self.bites as an int.
        """
        self.encode()
        return int.from_bytes(self.bites, byteorder="big")

    def hex(self):
        """
        hex returns self.bites as
        a hex string
        """
        return hex(self.int())

    def _encode_crc(self):
        """
        _encode_crc encode crc32
        """
        crc_int = crc32(self.bites)
        self.info_section.crc = hex(crc_int)
        self.bites += int.to_bytes(crc_int, 4, byteorder="big")

    def _unloop_descriptors(self):
        """
        _unloop_descriptors
        for each descriptor in self.descriptors
        encode descriptor tag, descriptor length,
        and the descriptor into all_bites.bites
        """
        all_bites = NBin()
        dbite_chunks = [dsptr.encode() for dsptr in self.descriptors]
        for chunk, dsptr in zip(dbite_chunks, self.descriptors):
            dsptr.descriptor_length = len(chunk)
            all_bites.add_int(dsptr.tag, 8)
            all_bites.add_int(dsptr.descriptor_length, 8)
            all_bites.add_bites(chunk)
        return all_bites.bites

    def load_info_section(self, gonzo):
        """
        load_info_section loads data for Cue.info_section
        isec should be a dict.
        if 'splice_command_type' is included,
        an empty command instance will be created for Cue.command
        """
        if "info_section" in gonzo:
            self.info_section.load(gonzo["info_section"])

    def load_command(self, gonzo):
        """
        load_command loads data for Cue.command
        cmd should be a dict.
        if 'command_type' is included,
        the command instance will be created.
        """
        if "command" not in gonzo:
            self.no_cmd()
        cmd = gonzo["command"]
        if "command_type" in cmd:
            self.command = command_map[cmd["command_type"]]()
            self.command.load(cmd)

    def load_descriptors(self, dlist):
        """
        Load_descriptors loads descriptor data.
        dlist is a list of dicts
        if 'tag' is included in each dict,
        a descriptor instance will be created.
        """
        if not isinstance(dlist, list):
            raise Exception("\033[7mdescriptors should be a list\033[27m")
        for dstuff in dlist:
            if "tag" in dstuff:
                dscptr = descriptor_map[dstuff["tag"]]()
                dscptr.load(dstuff)
                #  if dscptr.has('segmentation_upid'):
                #     dscptr.xml_redecode() # Expand upids
                self.descriptors.append(dscptr)

    def no_cmd(self):
        """
        no_cmd raises an exception if no splice command.
        """
        raise Exception("\033[7mA splice command is required\033[27m")

    def load(self, gonzo):
        """
        Cue.load loads SCTE35 data for encoding.
        gonzo is a dict or json
        with any or all of these keys
        gonzo = {
            'info_section': {dict} ,
            'command': {dict},
            'descriptors': [list of {dicts}],
            }
        """
        if isinstance(gonzo, bytes):
            gonzo = gonzo.decode()
        if isinstance(gonzo, str):
            try:
                gonzo = int(gonzo)
                self.bites = self._int_bits(int(gonzo))
                self.decode()
                return True
            except:
                if gonzo.strip()[0] == "<":
                    self.from_xml(gonzo)
                    return True

            gonzo = json.loads(gonzo)

        if "command" not in gonzo:
            self.no_cmd()
        self.load_info_section(gonzo)
        self.load_command(gonzo)
        self.load_descriptors(gonzo["descriptors"])
        self.encode()
        return True

    def from_xml(self, gonzo):
        """
        build_cue takes the data put into the gonzo dict
        and builds a fu.Cue instance
        """
        sxp = SuperXmlParser()
        dat = sxp.xml2cue(gonzo)
        if isinstance(
            dat, str
        ):  # a string  is returned for Binary xml tag, make sense?
            self.bites = self._mk_bits(dat)
            self.decode()
        else:
            self.load(dat)  # a dict is returned plain xml.
            for d in self.descriptors:
                if d.has("segmentation_upid"):
                    d.xml_redecode()  # Expand upids
            # Self.encode() will calculate lengths and types and such
            self.encode()

    def _xml_mk_descriptor(self, sis, ns):
        """
        _mk_descriptor_xml make xml nodes for descriptors.
        """
        for d in self.descriptors:
            if d.has("segmentation_type_id") and d.segmentation_type_id in table22:
                comment = f"{table22[d.segmentation_type_id]}"
                sis.add_comment(comment)
            sis.add_child(d.xml(ns=ns))
        return sis

    def xml(self, ns="scte35"):
        """
        xml returns a fu.Node instance
        which can be edited as needed or printed.
        xmlbin
        """
        sis = self.info_section.xml(ns=ns)
        # if not self.command:
        #    raise Exception("\033[7mA Splice Command is Required\033[27m")
        cmd = self.command.xml(ns=ns)
        sis.add_child(cmd)
        sis = self._xml_mk_descriptor(sis, ns)
        sis.mk()
        return sis

    def xmlbin(self, ns="scte35"):
        """
        xml returns a fu.Node instance
        which can be edited as needed or printed.
        xmlbin
        """
        sig_attrs = {"xmlns": "https://scte.org/schemas/35"}
        sig_node = Node("Signal", attrs=sig_attrs, ns=ns)
        bin_node = Node("Binary", value=self.encode(), ns=ns)
        sig_node.add_child(bin_node)
        return sig_node



"""

streamtypes.py

    Used by fu.Stream
    On import stream_types generates a complete streamtype_map.

        from .streamtypes import streamtype_map

"""

streamtype_map = {
    0x00: "Reserved",
    0x01: "MPEG-1 video",
    0x02: "H.262  video",
    0x03: "MPEG-2 audio",
    0x04: "MPEG-2 halved sample rate audio",
    0x05: "MPEG-2 tabled data",
    0x06: "MPEG-2 binary data",
    0x07: "ISO/IEC 13522 (MHEG)",
    0x08: "H.222 and  DSM CC",
    0x09: "H.222 and auxiliary data",
    0x0A: "DSM CC multiprotocol encapsulation",
    0x0B: "DSM CC U-N messages",
    0x0C: "DSM CC stream descriptors",
    0x0D: "DSM CC tabled data",
    0x0E: "auxiliary data",
    0x0F: "ADTS AAC ",
    0x10: "MPEG-4 H.263 based video",
    0x11: "MPEG-4 LOAS multi-format framed audio",
    0x12: "MPEG-4 FlexMux",
    0x13: "MPEG-4 FlexMux",
    0x14: "DSM CC ",
    0x15: "ID3",
    0x16: "Sectioned metadata",
    0x17: "DSM CC Data Carousel metadata",
    0x18: "DSM CC Object Carousel metadata",
    0x19: "Synchronized Download Protocol metadata",
    0x1A: "IPMP",
    0x1B: "H.264",
    0x1C: "MPEG-4 raw audio)",
    0x1D: "MPEG-4 text",
    0x1E: "MPEG-4 auxiliary video",
    0x1F: "SVC (MPEG-4 AVC sub-bitstream)",
    0x20: "MVC (MPEG-4 AVC sub-bitstream)",
    0x21: "JPEG 2000 video",
    # 0x22 - 0x23:"Reserved.",
    0x24: " H.265",
    # 0x25 - 0x41:"Reserved.",
    0x42: "Chinese Video Standard",
    # 0x43 - 0x7e:"Reserved.",
    0x7F: " DRM Trash",
    0x80: "H.262 with DES-64-CBC ",
    0x81: "AC-3",
    0x82: "SCTE subtitle",
    0x83: "Dolby TrueHD audio ",
    0x84: "AC-3",
    0x85: "DTS 8 channel audio for Blu-ray",
    0x86: "SCTE-35",
    0x87: "AC-3 up to 16 channel audio for ATSC",
    # 0x88 - 0x89:"Privately defined.",
    0x90: "Blu-ray Presentation Graphic Stream (subtitling)",
    0x91: "ATSC DSM CC Network Resources table",
    # 0x92 - 0xBF:"Privately defined.",
    0xC0: "DigiCipher II text",
    0xC1: "AC-3  with AES-128-CBC ",
    0xC2: "ATSC DSM CC synchronous data",
    # 0xC3 - 0xCE:"Privately defined.",
    0xCF: "ADTS AAC with AES-128-CBC frame encryption",
    0xD0: "Privately defined.",
    0xD1: "BBC Dirac ",
    0xD2: "AVS2 ",
    0xD3: "AVS3 Audio",
    0xD4: "AVS3 Video ",
    # 0xD5 - 0xDA:"Privately defined.",
    0xDB: " H.264 with AES-128-CBC slice encryption",
    # 0xDC - 0xE9:"Privately defined.",
    0xEA: "Windows janky media v9",
    # 0xEB - 0xFB:"Privately defined.",
    0xFC: "KLV",
}


others = {
    "Reserved": list(range(0x22, 0x23))
    + list(range(0x25, 0x41))
    + list(range(0x43, 0x7E)),
    "Private": list(range(0x88, 0x89))
    + list(range(0x92, 0xBF))
    + list(range(0xC3, 0xCE))
    + list(range(0xD5, 0xDA))
    + list(range(0xDC, 0xE9))
    + list(range(0xEB, 0xFB))
    + list(range(0xFD, 0xFF)),
}


def hex_literal(integer):
    """
    Return hex literal
    """
    return int(hex(integer), base=16)


def add_stream_types(alist, streamtype):
    """
    add_stream_types dynamically
    adds a range of a stream type,
    like "Reserved",to streamtype_map
    """
    for i in alist:
        j = hex_literal(i)
        streamtype_map[j] = streamtype


def mk_streamtype_map(others):
    """
    mk_streamtype_map dynamically adds
    stream types to streamtype_map
    """
    for k, v in others.items():
        add_stream_types(v, k)
    keys = list(streamtype_map.keys())
    keys.sort()


mk_streamtype_map(others)



"""
crc.py  crc32 function for encoding.
"""

POLY = 0x104C11DB7
INIT_VALUE = 0xFFFFFFFF
ZERO = 0x0
ONE = 0x1
EIGHT = 0x8
TWENTY_FOUR = 0x18
THIRTY_TWO = 0x20
TWO_FIFTY_FIVE = 0xFF
TWO_FIFTY_SIX = 0x100


def _bytecrc(crc, poly):
    mask = ONE << (THIRTY_TWO - ONE)
    i = EIGHT
    while i:
        crc = (crc << ONE, crc << ONE ^ poly)[crc & mask != ZERO]
        i -= ONE
    return crc & INIT_VALUE


def _mk_table():
    mask = (ONE << THIRTY_TWO) - ONE
    poly = POLY & mask
    return [_bytecrc((i << TWENTY_FOUR), poly) for i in range(TWO_FIFTY_SIX)]


def crc32(data):
    """
    generate a 32 bit crc
    """

    table = _mk_table()
    crc = INIT_VALUE
    for bite in data:
        crc = table[bite ^ ((crc >> TWENTY_FOUR) & TWO_FIFTY_FIVE)] ^ (
            (crc << EIGHT) & (INIT_VALUE - TWO_FIFTY_FIVE)
        )
    return crc


def crc32hex(data):
    """
    crc32hex crc32 returned as hex
    """
    return hex(crc32(data))



"""
sixfix.py
"""

import io
import sys
from fu.crc import crc32
from fu.bitn import NBin
from fu.spare import print2
from fu.stream import Stream

fixme = []


def passed(cue):
    """
    passed a no-op function
    """
    global fixme
    fixme.append(int(cue.packet_data.pid, base=16))
    return cue


class PreFix(Stream):
    """
    PreFix is used to gather 06 Bin data pids with SCTE-35.
    """

    def decode(self, func=passed):
        super().decode(func=passed)
        global fixme
        fixme = list(set(fixme))
        print("fixing these pids", fixme)
        return fixme


class SixFix(Stream):
    """
    SixFix class
    fixes bin data streams with SCTE-35 to 0x86 SCTE-35 streams
    """

    CUEI_DESCRIPTOR = b"\x05\x04CUEI"

    def __init__(self, tsdata=None):
        super().__init__(tsdata)
        self.pmt_payload = None
        self.con_pids = set()
        self.out_file = "sixfixed-" + tsdata.rsplit("/")[-1]
        self.in_file = sys.stdin.buffer

    def _parse_by_pid(self, pkt, pid):
        if pid in self.pids.tables:
            self._parse_tables(pkt, pid)
        if pid in self.pids.pmt:
            if self.pmt_payload:
                pkt = pkt[:4] + self.pmt_payload
        return pkt

    def convert_pids(self):
        """
        convert_pids
        changes the stream type to 0x86 and replaces
        the existing PMT as it writes packets to the outfile
        """
        active = io.BytesIO()
        pkt_count = 0
        chunk_size = 2048
        if isinstance(self.out_file, str):
            self.out_file = open(self.out_file, "wb")
        with self.out_file as out_file:
            for pkt in self.iter_pkts():
                pid = self._parse_pid(pkt[1], pkt[2])
                pkt = self._parse_by_pid(pkt, pid)
                active.write(pkt)
                pkt_count = (pkt_count + 1) % chunk_size
                if not pkt_count:
                    out_file.write(active.getbuffer())
                    active = io.BytesIO()

    def _regen_pmt(self, n_seclen, pcr_pid, n_proginfolen, n_info_bites, n_streams):
        nbin = NBin()
        nbin.add_int(2, 8)  # 0x02
        nbin.add_int(1, 1)  # section Syntax indicator
        nbin.add_int(0, 1)  # 0
        nbin.add_int(3, 2)  # reserved
        nbin.add_int(n_seclen, 12)  # section length
        nbin.add_int(1, 16)  # program number
        nbin.add_int(3, 2)  # reserved
        nbin.add_int(0, 5)  # version
        nbin.add_int(1, 1)  # current_next_indicator
        nbin.add_int(0, 8)  # section number
        nbin.add_int(0, 8)  # last section number
        nbin.add_int(7, 3)  # res
        nbin.add_int(pcr_pid, 13)
        nbin.add_int(15, 4)  # res
        nbin.add_int(n_proginfolen, 12)
        nbin.add_bites(n_info_bites)
        nbin.add_bites(n_streams)
        a_crc = crc32(nbin.bites)
        nbin.add_int(a_crc, 32)
        n_payload = nbin.bites
        pad = 187 - (len(n_payload) + 4)
        pointer_field = b"\x00"
        if pad > 0:
            n_payload = pointer_field + n_payload + (b"\xff" * pad)
        self.pmt_payload = n_payload

    def _parse_pmt(self, pay, pid):
        """
        parse program maps for streams
        """
        pay = self._chk_partial(pay, pid, self._PMT_TID)
        if not pay:
            return False
        seclen = self._parse_length(pay[1], pay[2])
        n_seclen = seclen + 6
        if self._section_incomplete(pay, pid, seclen):
            return False
        program_number = self._parse_program(pay[3], pay[4])
        pcr_pid = self._parse_pid(pay[8], pay[9])
        self.pids.pcr.add(pcr_pid)
        self.maps.pid_prgm[pcr_pid] = program_number
        proginfolen = self._parse_length(pay[10], pay[11])
        idx = 12
        n_proginfolen = proginfolen + len(self.CUEI_DESCRIPTOR)
        end = idx + proginfolen
        info_bites = pay[idx:end]
        n_info_bites = info_bites + self.CUEI_DESCRIPTOR
        while idx < end:
            # d_type = pay[idx]
            idx += 1
            d_len = pay[idx]
            idx += 1
            # d_bytes = pay[idx - 2 : idx + d_len]
            idx += d_len
        si_len = seclen - 9
        si_len -= proginfolen
        n_streams = self._parse_program_streams(si_len, pay, idx, program_number)
        self._regen_pmt(n_seclen, pcr_pid, n_proginfolen, n_info_bites, n_streams)
        return True

    def _parse_program_streams(self, si_len, pay, idx, program_number):
        """
        parse the elementary streams
        from a program
        """
        chunk_size = 5
        end_idx = (idx + si_len) - 4
        start = idx
        while idx < end_idx:
            pay, stream_type, pid, ei_len = self._parse_stream_type(pay, idx)
            idx += chunk_size
            idx += ei_len
            self.maps.pid_prgm[pid] = program_number
            self._set_scte35_pids(pid, stream_type)
        streams = pay[start:end_idx]
        return streams

    def _parse_stream_type(self, pay, idx):
        """
        extract stream pid and type
        """
        npay = pay
        stream_type = pay[idx]
        el_pid = self._parse_pid(pay[idx + 1], pay[idx + 2])
        if el_pid in self.con_pids:
            if stream_type == 6:
                npay = pay[:idx] + b"\x86" + pay[idx + 1 :]
        ei_len = self._parse_length(pay[idx + 3], pay[idx + 4])
        return npay, stream_type, el_pid, ei_len


def sixfix(arg):
    """
    sixfix converts 0x6 bin data mpegts streams
    that contain SCTE-35 data to stream type 0x86
    """
    s1 = PreFix(arg)
    sixed = s1.decode(func=passed)
    global fixme
    fixme = []
    if not sixed:
        print2("No bin data SCTE-35 streams were found.")
        return
    s2 = SixFix(arg)
    s2.con_pids = sixed
    s2.convert_pids()
    print2(f'Wrote: sixfixed-{arg.rsplit("/")[-1]}')
    return


if __name__ == "__main__":
    sixfix(sys.argv[1])



"""
fu.base contains
the class SCTE35Base.
"""

import json
from .bitn import NBin
from .spare import print2
from .xml import Node


class SCTE35Base:
    """
    SCTE35Base is a base class for
    SpliceCommand and SpliceDescriptor classes
    """

    ROLLOVER = 8589934591

    def __repr__(self):
        return str(self.__dict__)

    @staticmethod
    def _chk_nbin(nbin):
        if not nbin:
            nbin = NBin()
        return nbin

    def _chk_var(self, var_type, nbin_method, var_name, bit_count):
        """
        _chk_var is used to check var values and types before encoding
        """
        var_value = self.__dict__[var_name]
        if var_value is None:
            err_mesg = f"\033[7m{var_name} is not set, it should be {bit_count} bit(s) and type {var_type}\033[27m"
            raise ValueError(err_mesg)
        if not isinstance(var_value, var_type):
            err_mesg = f' \033[7m{var_name} is "{var_value}", it should be type {var_type}\033[27m and {bit_count} bit(s),\n '
            raise ValueError(err_mesg)
        nbin_method(var_value, bit_count)

    @staticmethod
    def as_90k(int_time):
        """
        ticks to 90k timestamps
        """
        return round((int_time / 90000.0), 6)

    @staticmethod
    def as_ticks(float_time):
        """
        90k timestamps to ticks
        """
        return int(round(float_time * 90000))

    @staticmethod
    def as_hms(secs_of_time):
        """
        as_hms converts timestamp to
        00:00:00.000 format
        """
        hours, seconds = divmod(secs_of_time, 3600)
        mins, seconds = divmod(seconds, 60)
        seconds = round(seconds, 3)
        output = f"{int(hours):02}:{int(mins):02}:{seconds:02}"
        if len(output.split(".")[1]) < 2:
            output += "0"
        return output

    @staticmethod
    def fix_hex(hexed):
        """
        fix_hex adds padded zero if needed for byte conversion.
        """
        return (hexed.replace("0x", "0x0", 1), hexed)[len(hexed) % 2 == 0]

    def get(self):
        """
        Returns instance as a kv_clean'ed dict
        """
        return self.kv_clean()

    def has(self, what):
        """
        has runs hasattr with self and what
        """
        if hasattr(self, what):
            if vars(self)[what]:
                return True
        return False

    @staticmethod
    def idxsplit(gonzo, sep):
        """
        idxsplit is like split but you keep
        the sep
        example:
                >>> idxsplit('123456789',4)
                >>>'456789'
        """
        if sep in gonzo:
            return gonzo[gonzo.index(sep) :]
        return False

    def json(self):
        """
        json returns self as kv_clean'ed json
        """
        return json.dumps(self.get(), indent=4)

    def kv_clean(self):
        """
        kv_clean recursively removes items
        from a dict if the value is None.
        """

        def b2l(val):
            if isinstance(val, (SCTE35Base)):
                val.kv_clean()
            if isinstance(val, (list)):
                val = [b2l(v) for v in val]
            if isinstance(val, (dict)):
                val = {k: b2l(v) for k, v in val.items()}
            if isinstance(val, (bytes, bytearray)):
                val = list(val)
            return val

        return {k: b2l(v) for k, v in vars(self).items() if v is not None}

    def load(self, gonzo):
        """
        load is used to load
        data from a dict or json string.
        only updates vars that exist in the obj.
        """
        if isinstance(gonzo, str):
            gonzo = json.loads(gonzo)
        if isinstance(gonzo, dict):
            prevars = vars(self)
            for k, v in gonzo.items():
                if k in prevars:
                    self.__dict__[k] = v

    def show(self):
        """
        show prints self as json to stderr (2)
        """
        print2(self.json())

    def xml(self, ns="scte35"):
        """
        xml default xml method will return
        all attributes, which is expressly allowed
        for SCTE-35 xml.
        """
        xml_attrs = self.kv_clean()
        iam = type(self)
        iam = str(iam).split("'", 1)[1].split("'", 1)[0]
        this = Node(iam, attrs=xml_attrs, ns=ns)
        return this



"""
fu/upids.py

classy Upids

"""

from .bitn import BitBin, NBin
from .xml import Node

charset = "ascii"  # this isn't a constant pylint.

"""
set charset to None to return raw bytes
"""


class Upid:
    """
    Upid base class handles URI UPIDS
    """

    def __init__(self, bitbin=None, upid_type=0, upid_length=0):
        self.bitbin = bitbin
        self.upid_type = upid_type
        self.upid_name = upid_map[upid_type][0]
        self.upid_length = upid_length
        self.bit_length = upid_length << 3
        self.upid_value = None

    def decode(self):
        """
        decode Upid
        """
        self.upid_value = self.bitbin.as_charset(self.bit_length, charset)
        return self.upid_name, self.upid_value

    def encode(self, nbin, seg_upid):
        """
        encode Upid
        """
        self.upid_value = seg_upid.encode()
        nbin.add_bites(self.upid_value)

    def redecode(self, seg_upid):
        """
        redecode is for decoding complex xml upids
        before encodingto another format.
        """
        if isinstance(seg_upid, str):
            try:
                self.bitbin = BitBin(bytes.fromhex(seg_upid))
            except ValueError:
                self.bitbin = BitBin(seg_upid.encode())
            return self.decode()
        else:
            self.upid_value = seg_upid
        return self.upid_name, self.upid_value

    def _xml_format_attr(self):
        """
        _xml_format_attr sets segmentation_upid_format
        """
        if self.upid_type in [0x03, 0x07, 0x09, 0x0E]:
            return "text"
        return "hexbinary"

    def xml(self, ns="scte35"):
        """
        xml return a upid xml node
        """
        ud_attrs = {
            "segmentation_upid_type": self.upid_type,
            "segmentation_upid_format": self._xml_format_attr(),
        }
        return Node("SegmentationUpid", attrs=ud_attrs, value=self.upid_value, ns=ns)

    def _extra_xml_attrs(self, ud):
        if "format_identifier" in self.upid_value:
            ud["format_identifier"] = int.from_bytes(
                self.upid_value["format_identifier"].encode(), byteorder="big"
            )
        if "private_data" in self.upid_value:
            ud["private_data"] = self.upid_value["private_data"]
        return ud

    def complexml(self, ns="scte35"):
        """
        complexml return a upid xml node for complex upids
        """
        ud_attrs = {
            "segmentation_upid_type": self.upid_type,
            "segmentation_upid_format": self._xml_format_attr(),
        }
        if isinstance(self.upid_value, dict):
            ud_attrs = self._extra_xml_attrs(ud_attrs)
        nbin = NBin()
        self.encode(nbin, self.upid_value)
        return Node("SegmentationUpid", attrs=ud_attrs, value=nbin.bites.hex(), ns=ns)


class NoUpid(Upid):
    """
    NoUpid class
    """

    def decode(self):
        """
        decode for no upid
        """
        return self.upid_name, "No UPID"

    def encode(self, nbin, seg_upid):
        """
        encode for no upid
        """
        nbin.forward(0)

    def xml(self, ns="scte35"):
        """
        xml return a upid xml node
        """
        ud_attrs = {
            "segmentation_upid_type": 0x0,
            "segmentation_upid_format": "hexbinary",
        }
        return Node("SegmentationUpid", attrs=ud_attrs, value="", ns=ns)


class AirId(Upid):
    """
    Air Id Upid
    """

    def decode(self):
        """
        decode AirId
        """
        self.upid_value = self.bitbin.as_hex(self.bit_length)
        return self.upid_name, self.upid_value

    def encode(self, nbin, seg_upid):
        """
        encode AirId
        """
        nbin.add_hex(seg_upid, (self.upid_length << 3))


class Atsc(Upid):
    """
    ATSC Upid
    """

    def decode(self):
        """
        decode Atsc Upid
        """
        cont_size = self.bit_length - 32
        self.upid_value = {
            "TSID": self.bitbin.as_int(16),
            "reserved": self.bitbin.as_int(2),
            "end_of_day": self.bitbin.as_int(5),
            "unique_for": self.bitbin.as_int(9),
            "content_id": self.bitbin.as_charset(cont_size, charset),
        }
        return self.upid_name, self.upid_value

    def encode(self, nbin, seg_upid):
        """
        encode Atsc
        """
        _, self.upid_value = self.redecode(seg_upid)
        nbin.add_int(self.upid_value["TSID"], 16)
        nbin.add_int(self.upid_value["reserved"], 2)
        nbin.add_int(self.upid_value["end_of_day"], 5)
        nbin.add_int(self.upid_value["unique_for"], 9)
        nbin.add_bites(self.upid_value["content_id"].encode("utf-8"))

    def xml(self, ns="scte35"):
        """
        xml return xml node
        """
        return self.complexml(ns=ns)


class Eidr(Upid):
    """
    Eidr Upid
    """

    def decode(self):
        """
        decode Eidr Upid
        """
        pre = self.bitbin.as_hex(16)
        post = []
        # switch to compact binary format
        nibbles = 20
        while nibbles:
            post.append(hex(self.bitbin.as_int(4))[2:])
            nibbles -= 1
        self.upid_value = f"{pre}{''.join(post)}"
        return self.upid_name, self.upid_value

    def encode(self, nbin, seg_upid):
        """
        encode Eidr Upid
        """
        _, self.upid_value = self.redecode(seg_upid)
        # switch to compact binary format
        nbin.add_hex(self.upid_value[:6], 16)
        substring = self.upid_value[6:]
        for i in substring:
            hexed = f"0x{i}"
            nbin.add_hex(hexed, 4)

    def xml(self, ns="scte35"):
        """
        xml return xml node
        """
        return self.complexml(ns=ns)


class Isan(Upid):
    """
    Isan Upid
    """

    def decode(self):
        """
        decode Isan Upid
        """
        self.upid_value = self.bitbin.as_hex(self.bit_length)
        return self.upid_name, self.upid_value

    def encode(self, nbin, seg_upid):
        """
        encode Isan Upid
        """
        self.upid_value = seg_upid
        nbin.add_hex(self.upid_value, (self.upid_length << 3))


class Mid(Upid):
    """
    Mid Upid
    """

    def decode(self):
        """
        decode Mid Upid
        """
        self.upid_value = []
        ulb = self.bit_length
        while ulb:
            upid_type = self.bitbin.as_int(8)  # 1 byte
            ulb -= 8
            upid_length = self.bitbin.as_int(8)
            ulb -= 8
            upid_type_name, segmentation_upid = upid_map[upid_type][1](
                self.bitbin, upid_type, upid_length
            ).decode()
            mid_upid = {
                "upid_type": upid_type,
                "upid_type_name": upid_type_name,
                "upid_length": upid_length,
                "segmentation_upid": segmentation_upid,
            }
            ulb -= upid_length << 3
            self.upid_value.append(mid_upid)
        return self.upid_name, self.upid_value

    def encode(self, nbin, seg_upid):
        """
        encode Mid Upid
        """
        for mid_upid in seg_upid:
            nbin.add_int(mid_upid["upid_type"], 8)
            nbin.add_int(mid_upid["upid_length"], 8)
            the_upid = upid_map[mid_upid["upid_type"]][1](
                None, mid_upid["upid_type"], mid_upid["upid_length"]
            )
            the_upid.encode(nbin, mid_upid["segmentation_upid"])

    def xml(self, ns="scte35"):
        """
        xml return a upid xml node

        """
        mid_nodes = []
        for u in self.upid_value:
            u_attrs = {
                "segmentation_upid_type": u["upid_type"],
                "name": u["upid_type_name"],
            }
            value = u["segmentation_upid"]
            node = Node("SegmentationUpid", attrs=u_attrs, value=value, ns=ns)
            mid_nodes.append(node)
        return mid_nodes


class Mpu(Upid):
    """
    Mpu Upid
    """

    def _decode_adfr(self):
        """
        decode_adfr handles Addressabkle TV MPU Upids
        """
        data = bytes.fromhex(self.upid_value["private_data"][2:])
        self.upid_value["version"] = data[0]
        self.upid_value["channel_identifier"] = hex(
            int.from_bytes(data[1:3], byteorder="big")
        )
        self.upid_value["date"] = int.from_bytes(data[3:7], byteorder="big")
        self.upid_value["break_code"] = int.from_bytes(data[7:9], byteorder="big")
        self.upid_value["duration"] = hex(int.from_bytes(data[9:11], byteorder="big"))

    def decode(self):
        """
        decode MPU Upids
        """
        self.upid_value = {
            "format_identifier": self.bitbin.as_charset(32),
            "private_data": self.bitbin.as_hex(self.bit_length - 32),
        }
        if self.upid_value["format_identifier"] == "ADFR":
            self._decode_adfr()
        return self.upid_name, self.upid_value

    def encode(self, nbin, seg_upid):
        """
        encode MPU Upids
        """
        _, self.upid_value = self.redecode(seg_upid)
        bit_len = self.bit_length
        nbin.add_bites(bytes(self.upid_value["format_identifier"], "utf8"))
        bit_len -= 32
        nbin.add_hex(self.upid_value["private_data"], bit_len)

    def xml(self, ns="scte35"):
        """
        xml return xml node
        """
        return self.complexml(ns=ns)


class Umid(Upid):
    """
    Umid Upid
    """

    def decode(self):
        """
        decode Umid Upids
        """
        chunks = []
        ulb = self.bit_length
        while ulb > 32:
            chunks.append(self.bitbin.as_hex(32)[2:])
            ulb -= 32
        self.upid_value = ".".join(chunks)
        return self.upid_name, self.upid_value

    def encode(self, nbin, seg_upid):
        """
        encode Umid Upid
        """
        _, self.upid_value = self.redecode(seg_upid)
        chunks = self.upid_value.split(".")
        for chunk in chunks:
            nbin.add_hex(chunk, 32)

    def xml(self, ns="scte35"):
        """
        xml return xml node
        """
        return self.complexml(ns=ns)


# segmentation_upid_type : [name, class, length]
upid_map = {
    0x00: ["No UPID", NoUpid, 0],
    0x01: [
        "Type 0x01 is deprecated. Use type 0x0C, MPU.",
        Upid,
        False,
    ],
    0x02: ["Type 0x02 is deprecated. Use type 0x03, AdID.", Upid, 8],
    0x03: ["AdID", Upid, 12],
    0x04: ["UMID", Umid, 32],
    0x05: ["Type 0x05 is deprecated. Use type 0x06, ISAN.", Isan, 8],
    0x06: ["ISAN", Isan, 12],
    0x07: ["TID", Upid, 12],
    0x08: ["AiringID", AirId, 8],
    0x09: ["ADI", Upid, False],
    0x0A: ["EIDR", Eidr, 12],
    0x0B: ["ATSC", Atsc, False],
    0x0C: ["MPU", Mpu, False],
    0x0D: ["MID", Mid, False],
    0x0E: ["ADS Info", Upid, False],
    0x0F: ["URI", Upid, False],
    0x10: ["UUID", Upid, 16],
    0x11: ["SCR", Upid, False],
    0xFD: ["Unknown", Upid, False],
}



"""
fu.__init__.py
"""

from .spare import print2
from .cue import Cue
from .section import SpliceInfoSection
from .segment import Segment
from .stream import Stream
from .version import version

from .commands import (
    TimeSignal,
    SpliceInsert,
    SpliceNull,
    PrivateCommand,
    BandwidthReservation,
)

from .descriptors import (
    AvailDescriptor,
    DVBDASDescriptor,
    DtmfDescriptor,
    SegmentationDescriptor,
    SpliceDescriptor,
    TimeDescriptor,
)



"""
xml.py  The Node class for converting to xml,
        The NameSpace Class.
        and several helper functions
"""

from xml.sax.saxutils import escape, unescape


def t2s(v):
    """
    _t2s converts
    90k ticks to seconds and
    rounds to six decimal places
    """
    return round(v / 90000.0, 6)


def un_camel(k):
    """
    camel changes camel case xml names
    to underscore_format names.
    """
    k = strip_ns(k)
    k = "".join([f"_{i.lower()}" if i.isupper() else i for i in k])
    return (k, k[1:])[k[0] == "_"]


def un_xml(v):
    """
    un_xml converts an xml value
    to ints, floats and booleans.
    """
    mapped = {
        "false": False,
        "true": True,
    }
    if v.isdigit():
        return int(v)
    if v.replace(".", "").isdigit():
        return float(v)
    if v in mapped:
        return mapped[v]
    return v


def strip_ns(this):
    """
    strip_ns strip namespace off this.
    """
    if "xmlns:" in this:
        return "xmlns"
    return this.split(":")[-1]


def iter_attrs(attrs):
    """
    iter_attrs normalizes xml attributes
    and adds them to the gonzo dict.
    """
    conv = {un_camel(k): un_xml(v) for k, v in attrs.items()}
    pts_vars = ["pts_time", "pts_adjustment", "duration", "segmentation_duration"]
    return {k: (t2s(v) if k in pts_vars else v) for k, v in conv.items()}


def val2xml(val):
    """
    val2xmlconvert val for xml
    """
    if isinstance(val, (bool, int, float)):
        return str(val).lower()
    if isinstance(val, str):
        if val.lower()[:2] == "0x":
            return str(int(val, 16))
    return val


def key2xml(string):
    """
    key2xml convert name to camel case
    """
    new_string = string
    if "_" in string:
        new_string = string.title().replace("_", "")
    return new_string[0].lower() + new_string[1:]


def mk_xml_attrs(attrs):
    """
    mk_xml_attrs converts a dict into
    a dict of xml friendly keys and values
    """
    return "".join([f' {key2xml(k)}="{val2xml(v)}"' for k, v in attrs.items()])


class NameSpace:
    """
    Each Node instance has a NameSpace instance
    to track namespace settings.
    @ns is the name of the namespace
    @uri is the xmlns uri
    @all is a flag to signal that all elements and
    attributes will have the namespace prefixed.
    By default only  the elements are prefixed with the namespace.
    """

    def __init__(self, ns=None, uri=None):
        self.ns = ns
        self.uri = uri
        self.all = False

    def prefix_all(self, abool=True):
        """
        prefix_all takes a boolean
        True turns it on, False turns it off.
        """
        self.all = abool

    def xmlns(self):
        """
        xmlns return xmlns attribute
        """
        if not self.uri:
            return ""
        if not self.ns:
            return f'xmlns="{self.uri}"'
        return f'xmlns:{self.ns}="{self.uri}"'

    def clear(self):
        """
        clear clear namespace info
        """
        self.ns = None
        self.uri = None
        self.all = False


class Node:
    """
    The Node class is to create an xml node.

    An instance of Node has:

        name :      <name> </name>
        value  :    <name>value</name>
        attrs :     <name attrs[k]="attrs[v]">
        children :  <name><children[0]></children[0]</name>
        depth:      tab depth for printing (automatically set)

    Use like this:

        from fu.xml import Node

        ts = Node('TimeSignal')
        st = Node('SpliceTime',attrs={'pts_time':3442857000})
        ts.add_child(st)
        print(ts)
    """

    def __init__(self, name, value="", attrs=None, ns=None):
        self.name = name
        self.value = value
        self.depth = 0
        self.namespace = NameSpace()
        self.namespace.ns = ns
        self.attr = None
        self._handle_attrs(attrs)
        self.children = []

    def __repr__(self):
        return self.mk()

    def _handle_attrs(self, attrs):
        if not attrs:
            attrs = {}
        if "xmlns" in attrs:
            self.namespace.uri = attrs.pop("xmlns")
        self.attrs = attrs

    def attrs2nodes(self):
        """
        attrs2nodes attributes to elements
        """
        for k, v in self.attrs.items():
            self.add_child(Node(name=key2xml(k), value=val2xml(v)), slot=0)
        self.attrs = {}

    def mk_ans(self, attrs):
        """
        mk_ans set namespace on attributes
        """
        new_attrs = {}
        if self.namespace.all:
            for k, v in attrs.items():
                new_attrs[f"{self.namespace.ns}:{k}"] = v
        return new_attrs

    def chk_obj(self, obj):
        """
        chk_obj determines if
        obj is self, or another obj
        for self.set_ns and self.mk
        """
        if obj is None:
            obj = self
        return obj

    def set_ns(self, ns=None):
        """
        set_ns set namespace on the Node
        """
        self.namespace.ns = ns

    def rm_attr(self, attr):
        """
        rm_attr remove an attribute
        """
        self.attrs.pop(attr)

    def add_attr(self, attr, value):
        """
        add_attr add an attribute
        """
        self.attrs[attr] = value

    def set_depth(self):
        """
        set_depth is used to format
        tabs in output
        """
        for child in self.children:
            child.depth = self.depth + 1

    def get_indent(self):
        """
        get_indent returns a string of spaces the required depth for a node
        """
        tab = "   "
        return tab * self.depth

    def _rendrd_children(self, rendrd, ndent, name):
        for child in self.children:
            rendrd += self.mk(child)
        return f"{rendrd}{ndent}</{name}>\n".replace(" >", ">")

    def mk_name(self):
        """
        mk_name add namespace to node name
        """
        name = self.name
        if self.namespace.ns:
            name = f"{self.namespace.ns}:{name}"
        return name

    def rendr_attrs(self, ndent, name):
        """
        rendrd_attrs renders xml attributes
        """
        attrs = self.attrs
        if self.namespace.all:
            attrs = self.mk_ans(self.attrs)
        new_attrs = mk_xml_attrs(attrs)
        if self.depth == 0:
            return f"{ndent}<{name} {self.namespace.xmlns()} {new_attrs}>"
        return f"{ndent}<{name}{new_attrs}>"

    def children_namespaces(self):
        """
        children_namespaces give children your namespace
        """
        for child in self.children:
            child.namespace.ns = self.namespace.ns
            child.namespace.all = self.namespace.all
            child.namespace.uri = ""

    def rendr_all(self, ndent, name):
        """
        rendr_all renders the Node instance and it's children in xml.
        """
        rendrd = self.rendr_attrs(ndent, name)
        if self.value:
            return f"{rendrd}{self.value}</{name}>\n"
        rendrd = f"{rendrd}\n"
        rendrd.replace(" >", ">")
        if self.children:
            return self._rendrd_children(rendrd, ndent, name)
        return rendrd.replace(">", "/>")

    def mk(self, obj=None):
        """
        mk make the Node as xml.
        """
        obj = self.chk_obj(obj)
        obj.set_depth()
        obj.children_namespaces()
        name = obj.mk_name()
        ndent = obj.get_indent()
        # obj.attrs2nodes()
        if isinstance(obj, Comment):
            return obj.mk(obj)
        return obj.rendr_all(ndent, name)

    def add_child(self, child, slot=None):
        """
        add_child adds a child node
        set slot to insert at index slot.
        """
        if not slot:
            slot = len(self.children)
        self.children = self.children[:slot] + [child] + self.children[slot:]

    def rm_child(self, child):
        """
        rm_child remove a child

        example:
        a_node.rm_child(a_node.children[3])
        """
        self.children.remove(child)

    def add_comment(self, comment, slot=None):
        """
        add_comment add a Comment node
        """
        self.add_child(Comment(comment), slot)


class Comment(Node):
    """
    The Comment class is to create a Node representing a xml comment.

    An instance of Comment has:

        name :      <!-- name -->
        depth:      tab depth for printing (automatically set)

    Since Comment is a Node, it also has attrs, value and children but
    these are ignored. cf etree.Comment
    Use like this:

        from fu.xml import Comment, Node

        n = Node('root')
        c = Comment('my first comment')

        n.add_child(c)
        print(n)

    See also Node.add_comment:
    """

    def mk(self, obj=None):
        obj = self.chk_obj(obj)
        obj.set_depth()
        return f"{obj.get_indent()}<!-- {obj.name} -->\n"



"""
SCTE35 Splice Descriptors
"""

from .bitn import BitBin
from .base import SCTE35Base
from .segmentation import table20, table22, dvb_table2
from .upids import upid_map
from .xml import Node


def k_by_v(adict, avalue):
    """
    dict key lookup by value
    """
    flipped = {v: k for k, v in adict.items()}
    if avalue in flipped:
        return flipped[avalue]
    return None


class SpliceDescriptor(SCTE35Base):
    """
    SpliceDescriptor is the
    base class for all splice descriptors.
    It should not be used directly
    """

    def __init__(self, bites=None):

        self.tag = None
        self.identifier = "CUEI"
        self.name = None
        self.bites = bites
        self.parse_tag_and_len()
        self.parse_id()
        self.private_data = None

    def parse_tag_and_len(self):
        """
        parse_tag_and_len
        parses the descriptors tag and length
        from self.bites
        """
        if self.bites:
            self.tag = self.bites[0]
            self.descriptor_length = self.bites[1]
            self.bites = self.bites[2:]

    def parse_id(self):
        """
        parse splice descriptor identifier
        """
        if self.bites:
            self.identifier = self.bites[:4].decode()
            # disabled for ffmv30
            #      if self.identifier != "CUEI":
            #          raise Exception('Identifier Is Not "CUEI"')
            self.bites = self.bites[4:]

    def decode(self):
        """
        decode handles Private Descriptors
        """
        self.private_data = self.bites.decode()
        if isinstance(self.private_data, bytes):
            self.private_data = self.bites.decode()

    def encode(self, nbin=None):
        """
        SpliceDescriptor.encode
        """
        nbin = self._chk_nbin(nbin)
        self._encode_id(nbin)
        if self.private_data:
            if isinstance(self.private_data, str):
                self.private_data = self.private_data.encode()
            nbin.add_bites(self.private_data)
        if self.tag in descriptor_map:
            return nbin
        return nbin.bites

    def _encode_id(self, nbin):
        """
        parse splice descriptor identifier
        """
        # self.identifier = "CUEI"
        id_int = int.from_bytes(self.identifier.encode(), byteorder="big")
        nbin.add_int(id_int, 32)


class DVBDASDescriptor(SpliceDescriptor):
    """
    Experimental DVB Descriptor Support
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.tag = 240  # 0xf0
        self.name = "DVD DAS Descriptor"
        self.identifier = "DVB_"
        self.break_num = 0
        self.breaks_expected = 0
        self.equivalent_segmentation_type = None
        self.equivalent_segmentation_message = None
        self.upid_type = 0x0F
        self.upid_type_name = None
        self.upid_length = None
        self.upid = None

    def decode(self):
        """
        Decode DVB DAS Descriptor
        """
        bitbin = BitBin(self.bites)
        self.break_num = bitbin.as_int(8)
        self.breaks_expected = bitbin.as_int(8)
        bitbin.forward(4)
        self.equivalent_segmentation_type = bitbin.as_int(4)
        if self.equivalent_segmentation_type in dvb_table2:
            self.equivalent_segmentation_message = dvb_table2[
                self.equivalent_segmentation_type
            ]
        self.upid_length = len(self.bites) - 3
        the_upid = upid_map[self.upid_type][1](bitbin, self.upid_type, self.upid_length)
        self.upid_type_name, self.upid = the_upid.decode()

    def encode(self, nbin=None):
        """
        encode DVB DAS Descriptor
        """
        nbin = super().encode(nbin)
        nbin.add_int(self.break_num, 8)
        nbin.add_int(self.breaks_expected, 8)
        nbin.forward(4)
        nbin.add_int(self.equivalent_segmentation_type, 4)
        the_upid = upid_map[self.upid_type][1](None, self.upid_type, self.upid_length)
        the_upid.encode(nbin, self.upid)
        return nbin.bites


class AvailDescriptor(SpliceDescriptor):
    """
    Table 17 -  avail_descriptor()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.name = "Avail Descriptor"
        self.tag = 0
        self.provider_avail_id = None

    def decode(self):
        """
        decode SCTE35 Avail Descriptor
        """
        bitbin = BitBin(self.bites)
        self.provider_avail_id = bitbin.as_int(32)

    def encode(self, nbin=None):
        """
        encode SCTE35 Avail Descriptor
        """
        nbin = super().encode(nbin)
        self._chk_var(int, nbin.add_int, "provider_avail_id", 32)
        return nbin.bites

    def xml(self, ns="scte35"):
        """
        Create a Node describing the AvailDescriptor
        """
        ad = Node(
            "AvailDescriptor", attrs={"providerAvailId": self.provider_avail_id}, ns=ns
        )
        return ad


class DtmfDescriptor(SpliceDescriptor):
    """
    Table 18 -  DTMF_descriptor()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.name = "DTMF Descriptor"
        self.tag = 1
        self.preroll = None
        self.dtmf_count = 0
        self.dtmf_chars = None

    def decode(self):
        """
        decode SCTE35 Dtmf Descriptor
        """
        self.preroll = self.bites[0]
        self.dtmf_count = self.bites[1] >> 5
        self.bites = self.bites[2:]
        self.dtmf_chars = list(self.bites[: self.dtmf_count].decode())

    def encode(self, nbin=None):
        """
        encode SCTE35 Dtmf Descriptor
        """
        nbin = super().encode(nbin)
        self._chk_var(int, nbin.add_int, "preroll", 8)
        d_c = 0
        self._chk_var(int, nbin.add_int, "dtmf_count", 3)
        nbin.forward(5)
        while d_c < self.dtmf_count:
            nbin.add_int(ord(self.dtmf_chars[d_c]), 8)
            d_c += 1
        return nbin.bites

    def xml(self, ns="scte35"):
        """
        Create a Node describing a DTMFDescriptor
        """
        dd = Node(
            "DTMFDescriptor",
            attrs={
                "preroll": self.preroll,
                "chars": "".join(self.dtmf_chars),
            },
            ns=ns,
        )
        return dd


##    def from_xml(self, gonzo):
##        """
##        Load an DTMFDescriptor from XML
##        """
##        if "DTMFDescriptor" in gonzo:
##            gonzo["DTMFDescriptor"]["dtmf_chars"] = gonzo["DTMFDescriptor"].pop("chars")
##            self.load(gonzo["DTMFDescriptor"])
##            self.dtmf_count = len(self.dtmf_chars)


class TimeDescriptor(SpliceDescriptor):
    """
    Table 25 - time_descriptor()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.tag = 3
        self.name = "Time Descriptor"
        self.tai_seconds = 0
        self.tai_ns = 0
        self.utc_offset = 0

    def decode(self):
        """
        decode SCTE35 Time Descriptor
        """
        bitbin = BitBin(self.bites)
        self.tai_seconds = bitbin.as_int(48)
        self.tai_ns = bitbin.as_int(32)
        self.utc_offset = bitbin.as_int(16)

    def encode(self, nbin=None):
        """
        encode SCTE35 Time Descriptor
        """
        nbin = super().encode(nbin)
        self._chk_var(int, nbin.add_int, "tai_seconds", 48)
        self._chk_var(int, nbin.add_int, "tai_ns", 32)
        self._chk_var(int, nbin.add_int, "utc_offset", 16)
        return nbin.bites

    def xml(self, ns="scte35"):
        """
        create a Node describing a TimeDescriptor
        """
        td = Node(
            "TimeDescriptor",
            attrs={
                "tai_seconds": self.tai_seconds,
                "tai_ns": self.tai_ns,
                "utc_offset": self.utc_offset,
            },
            ns=ns,
        )
        return td


##    def from_xml(self, gonzo):
##        """
##        load a TimeDescriptor from XML
##        """
##        if "TimeDescriptor" in gonzo:
##            self.load(gonzo["TimeDescriptor"])


class SegmentationDescriptor(SpliceDescriptor):
    """
    Table 19 - segmentation_descriptor()
    """

    SUB_SEG_TYPES = [
        0x30,
        0x32,
        0x34,
        0x36,
        0x38,
        0x3A,
        0x44,
        0x46,
    ]

    def __init__(self, bites=None):
        super().__init__(bites)
        self.tag = 2
        self.name = "Segmentation Descriptor"
        self.segmentation_event_cancel_indicator = None
        self.segmentation_event_id = None
        self.segmentation_event_id_compliance_indicator = None
        self.program_segmentation_flag = None
        self.segmentation_duration_flag = None
        self.delivery_not_restricted_flag = None
        self.web_delivery_allowed_flag = None
        self.no_regional_blackout_flag = None
        self.archive_allowed_flag = None
        self.device_restrictions = None
        self.segmentation_duration = None
        self.segmentation_message = None
        self.segmentation_type_id = None
        self.segmentation_upid_length = None
        self.segmentation_upid_type = None
        self.segmentation_upid_type_name = None
        self.segmentation_upid = None
        self.segment_num = None
        self.segments_expected = None
        self.sub_segment_num = None
        self.sub_segments_expected = None

    def decode(self):
        """
        decode a segmentation descriptor
        """
        bitbin = BitBin(self.bites)
        self.segmentation_event_id = bitbin.as_hex(32)
        self.segmentation_event_cancel_indicator = bitbin.as_flag(1)
        self.segmentation_event_id_compliance_indicator = bitbin.as_flag(1)
        bitbin.forward(6)
        if not self.segmentation_event_cancel_indicator:
            self._decode_flags(bitbin)
            self._decode_segmentation(bitbin)

    def _decode_flags(self, bitbin):
        self.program_segmentation_flag = bitbin.as_flag(1)
        self.segmentation_duration_flag = bitbin.as_flag(1)
        self.delivery_not_restricted_flag = bitbin.as_flag(1)
        if not self.delivery_not_restricted_flag:
            self.web_delivery_allowed_flag = bitbin.as_flag(1)
            self.no_regional_blackout_flag = bitbin.as_flag(1)
            self.archive_allowed_flag = bitbin.as_flag(1)
            self.device_restrictions = table20[bitbin.as_int(2)]
        else:
            bitbin.forward(5)

    def _decode_segmentation(self, bitbin):
        if self.segmentation_duration_flag:
            segmentation_duration_ticks = bitbin.as_int(40)
            self.segmentation_duration = self.as_90k(segmentation_duration_ticks)
        self.segmentation_upid_type = bitbin.as_int(8)
        self.segmentation_upid_length = bitbin.as_int(8)
        the_upid = self.mk_the_upid(bitbin)
        self.segmentation_upid_type_name, self.segmentation_upid = the_upid.decode()
        self.segmentation_type_id = bitbin.as_int(8)
        if self.segmentation_type_id in table22:
            self.segmentation_message = table22[self.segmentation_type_id]
        self._decode_segments(bitbin)

    def _chk_sub_segments(self):
        """
        chk_sub_segments sets sub_segment vars if not present
        """
        if self.segmentation_type_id in self.SUB_SEG_TYPES:
            if not self.sub_segment_num:
                self.sub_segment_num = 0
                self.sub_segments_expected = 0

    def _decode_segments(self, bitbin):
        self.segment_num = bitbin.as_int(8)
        self.segments_expected = bitbin.as_int(8)
        if self.segmentation_type_id in self.SUB_SEG_TYPES:
            # if sub_segment_num and sub_segments_expected
            # are not available set both of them to zero
            try:
                self.sub_segment_num = bitbin.as_int(8)
                self.sub_segments_expected = bitbin.as_int(8)
            finally:
                self._chk_sub_segments()

    def encode(self, nbin=None):
        """
        encode a segmentation descriptor
        """
        nbin = super().encode(nbin)
        self._chk_var(str, nbin.add_hex, "segmentation_event_id", 32)
        self._chk_var(bool, nbin.add_flag, "segmentation_event_cancel_indicator", 1)
        self._chk_var(
            bool, nbin.add_flag, "segmentation_event_id_compliance_indicator", 1
        )
        nbin.forward(6)
        if not self.segmentation_event_cancel_indicator:
            self._encode_flags(nbin)

            self._encode_segmentation(nbin)
        return nbin.bites

    def _encode_flags(self, nbin):
        self._chk_var(bool, nbin.add_flag, "program_segmentation_flag", 1)
        self._chk_var(bool, nbin.add_flag, "segmentation_duration_flag", 1)
        self._chk_var(bool, nbin.add_flag, "delivery_not_restricted_flag", 1)
        if not self.delivery_not_restricted_flag:
            self._chk_var(bool, nbin.add_flag, "web_delivery_allowed_flag", 1)
            self._chk_var(bool, nbin.add_flag, "no_regional_blackout_flag", 1)
            self._chk_var(bool, nbin.add_flag, "archive_allowed_flag", 1)
            a_key = k_by_v(table20, self.device_restrictions)
            nbin.add_int(a_key, 2)
        else:
            nbin.reserve(5)

    def mk_the_upid(self, bitbin=None):
        """
        mk_the_upid create a upid instance
        and return it. the bitbin arg is only
        used in decode()
        """
        upid_type = self.segmentation_upid_type
        if upid_type not in upid_map:
            upid_type = 0xFD
        the_upid = upid_map[upid_type][1](
            bitbin, upid_type, self.segmentation_upid_length
        )
        return the_upid

    def _encode_segmentation(self, nbin):
        if self.segmentation_duration_flag:
            nbin.add_int(self.as_ticks(self.segmentation_duration), 40)
        self._chk_var(int, nbin.add_int, "segmentation_upid_type", 8)
        self._chk_var(int, nbin.add_int, "segmentation_upid_length", 8)
        upid_type = self.segmentation_upid_type
        if upid_type not in upid_map:
            upid_type = 0xFD
        the_upid = upid_map[upid_type][1](
            None, upid_type, self.segmentation_upid_length
        )
        the_upid.encode(nbin, self.segmentation_upid)
        self._chk_var(int, nbin.add_int, "segmentation_type_id", 8)
        self._encode_segments(nbin)

    def _encode_segments(self, nbin):
        self._chk_var(int, nbin.add_int, "segment_num", 8)
        self._chk_var(int, nbin.add_int, "segments_expected", 8)
        if self.segmentation_type_id in [
            0x30,
            0x32,
            0x34,
            0x36,
            0x38,
            0x3A,
            0x44,
            0x46,
        ]:
            try:
                self._chk_var(int, nbin.add_int, "sub_segment_num", 8)
                self._chk_var(int, nbin.add_int, "sub_segments_expected", 8)
            except:
                nbin.add_int(0, 8)
                nbin.add_int(0, 8)

    def _xml_sub_segs(self, sd_attrs):
        if self.segmentation_type_id in self.SUB_SEG_TYPES:
            sd_attrs["sub_segment_num"] = self.sub_segment_num
            sd_attrs["sub_segments_expected"] = self.sub_segments_expected
        if self.segmentation_duration_flag:
            sd_attrs["segmentation_duration"] = self.as_ticks(
                self.segmentation_duration
            )
        return sd_attrs

    def xml(self, ns="scte35"):
        """
        Create a Node describing a SegmentationDescriptor
        """
        sd_attrs = {
            "segmentation_event_id": int(self.segmentation_event_id, 0),
            "segmentation_event_cancel_indicator": self.segmentation_event_cancel_indicator,
            "segmentation_event_id_compliance_indicator": self.segmentation_event_id_compliance_indicator,
            "segmentation_type_id": self.segmentation_type_id,
            "segment_num": self.segment_num,
            "segments_expected": self.segments_expected,
        }
        sd_attrs = self._xml_sub_segs(sd_attrs)
        sd = Node("SegmentationDescriptor", attrs=sd_attrs, ns=ns)
        the_upid = self.mk_the_upid()
        the_upid.upid_value = self.segmentation_upid
        upid_node = the_upid.xml(ns=ns)
        if not self.delivery_not_restricted_flag:
            dr_attrs = {
                "web_delivery_allowed_flag": self.web_delivery_allowed_flag,
                "no_regional_blackout_flag": self.no_regional_blackout_flag,
                "archive_allowed_flag": self.archive_allowed_flag,
                "device_restrictions": k_by_v(table20, self.device_restrictions),
            }
            dr = Node("DeliveryRestrictions", attrs=dr_attrs, ns=ns)
            sd.add_child(dr)
        comment = f"{upid_map[self.segmentation_upid_type][0]}"
        sd.add_comment(comment)
        if isinstance(upid_node, list):
            for node in upid_node:
                sd.add_child(node)
        else:
            sd.add_child(upid_node)
        return sd

    def xml_redecode(self):
        """
        redecode is for decoding complex xml upids
        before encoding to another format.
        """
        seg_upid = self.segmentation_upid
        if isinstance(seg_upid, str):
            bites = b""
            bitbin = None
            try:
                bites = bytes.fromhex(seg_upid)
            except ValueError:
                bites = seg_upid.encode()
            bitbin = BitBin(bites)
            self.segmentation_upid_length = len(bites)
            the_upid = self.mk_the_upid(bitbin=bitbin)
            self.segmentation_upid_type_name, self.segmentation_upid = the_upid.decode()


# map of known descriptors and associated classes
descriptor_map = {
    0: AvailDescriptor,
    1: DtmfDescriptor,
    2: SegmentationDescriptor,
    3: TimeDescriptor,
    240: DVBDASDescriptor,
}


def splice_descriptor(bites):
    """
    replaced splice_descriptor
    """
    spliced = None
    tag = bites[0]
    if tag in descriptor_map:
        spliced = descriptor_map[tag](bites)
    else:
        spliced = SpliceDescriptor(bites)
    spliced.decode()
    return spliced



"""
Mpeg-TS Stream parsing class Stream
"""

import sys

from functools import partial
from new_reader import reader
from .spare import print2
from .cue import Cue
from .packetdata import PacketData
from .streamtypes import streamtype_map


def no_op(cue):
    """
    no_op is just a dummy func to pass to Stream.decode()
    to suppress output.
    """
    return cue


def show_cue(cue):
    """
    default function call for Stream.decode
    when a SCTE-35 packet is found.
    """
    cue.show()


def show_cue_stderr(cue):
    """
    print2 cue data to sys.stderr
    for Stream.decode_proxy
    """
    cue.to_stderr()


class ProgramInfo:
    """
    ProgramInfo is a class to
    hold Program information
    for use with Stream.show()
    """

    def __init__(self, pid=None, pcr_pid=None):
        self.pid = pid
        self.pcr_pid = pcr_pid
        self.provider = b""
        self.service = b""
        self.streams = {}  # pid to stream_type mapping

    def show(self):
        """
        show print2 the Program Infomation
        in a familiar format.
        """
        serv = self.service.decode(errors="ignore")
        prov = self.provider.decode(errors="ignore")
        print2("")
        print2(f"\tService:  {serv}")
        print2(f"\tProvider: {prov}")
        print2(f"\tPid:      {self.pid}")
        print2(f"\tPcr Pid:  {self.pcr_pid}")
        print2("\tStreams:")
        # sorted_dict = {k:my_dict[k] for k in sorted(my_dict)})
        keys = sorted(self.streams)
        print2("\t  Pid\t\tType")
        for k in keys:
            vee = int(self.streams[k], base=16)
            if vee in streamtype_map:
                vee = f"{hex(vee)}\t{streamtype_map[vee]}"
            else:
                vee = f"{vee} Unknown"
            print2(f"\t  {k} [{hex(k)}]\t{vee}")


class Pids:
    """
    Pids holds sets of pids for pat,pcr,pmt, and scte35
    """

    SDT_PID = 0x11
    PAT_PID = 0x00

    def __init__(self):
        self.pcr = set()
        self.pmt = set()
        self.scte35 = set()
        self.maybe_scte35 = set()
        self.tables = set()
        self.tables.add(self.PAT_PID)
        self.tables.add(self.SDT_PID)


class Maps:
    """
    Maps holds mappings
    pids mapped to continuity_counters,
    programs, partial tables and last payload.

    programs mapped to pcr and pts

    """

    def __init__(self):
        self.pid_cc = {}
        self.pid_prgm = {}
        self.prgm_pcr = {}
        self.prgm_pts = {}
        self.prgm = {}
        self.partial = {}
        self.last = {}


class Stream:
    """
    Stream class for parsing MPEG-TS data.
    """

    # the _CONST are deprecated
    # please switch to CONST

    _PACKET_SIZE = PACKET_SIZE = 188
    _SYNC_BYTE = SYNC_BYTE = 0x47

    # tids

    _PMT_TID = PMT_TID = b"\x02"
    _SCTE35_TID = SCTE35_TID = b"\xfc"
    _SDT_TID = SDT_TID = b"\x42"
    ROLLOVER = 8589934591  # 95443.717678
    ROLLOVER9K = 95443.717678
    SCTE35_PES_START = b"\x00\x00\x01\xfc"

    def __init__(self, tsdata, show_null=True):
        """
        tsdata is an file or http/https url
        set show_null=False to exclude Splice Nulls

        Use like...

        from fu import Stream
        strm = Stream("vid.ts",show_null=False)
        strm.decode()

        """
        if isinstance(tsdata, str):
            self._tsdata = reader(tsdata)
        else:
            self._tsdata = tsdata
        self.show_null = show_null
        self.start = {}
        self.info = False
        self.the_program = None
        self.the_scte35_pids = []
        self.pids = Pids()
        self.maps = Maps()

    def __repr__(self):
        return str(self.__dict__)

    @staticmethod
    def as_90k(ticks):
        """
        as_90k returns ticks as 90k clock time
        """
        return round((ticks / 90000.0), 6)

    @staticmethod
    def _pusi_flag(pkt):
        return pkt[1] & 0x40

    @staticmethod
    def _afc_flag(pkt3):
        return pkt3 & 0x20

    @staticmethod
    def _pcr_flag(pkt):
        return pkt[5] & 0x10

    @staticmethod
    def _spi_flag(pkt):
        return pkt[5] & 0x20

    @staticmethod
    def _pts_flag(pay):
        # uses pay not pkt
        return pay[7] & 0x80

    @staticmethod
    def _parse_length(byte1, byte2):
        """
        parse a 12 bit length value
        """
        return (byte1 & 0xF) << 8 | byte2

    @staticmethod
    def _parse_pid(byte1, byte2):
        """
        parse a 13 bit pid value
        """
        pid = (byte1 & 0x1F) << 8 | byte2
        return pid

    @staticmethod
    def _parse_program(byte1, byte2):
        """
        parse a 16 bit program number value
        """
        return (byte1 << 8) | byte2

    @staticmethod
    def _split_by_idx(pay, marker):
        try:
            return pay[pay.index(marker) :]
        except:
            return False

    def _find_start(self):
        """
        _find_start locates the first SYNC_BYTE
        parses 1 packet and returns True
        if SYNC_BYTE is found.
        """
        while self._tsdata:
            one = self._tsdata.read(1)
            if not one:
                print2("\nNo Stream Found. \n")
                return False
            if one[0] == self.SYNC_BYTE:
                tail = self._tsdata.read(self.PACKET_SIZE - 1)
                if tail:
                    self._parse(one + tail)
                    return True
        return False

    def iter_pkts(self, num_pkts=1):
        """
        iter_pkts iterates a mpegts stream into packets
        """
        if self._find_start():
            return iter(partial(self._tsdata.read, self.PACKET_SIZE * num_pkts), b"")
        return False

    def _mk_pkts(self, chunk):
        return [
            self._parse(chunk[i : i + self.PACKET_SIZE])
            for i in range(0, len(chunk), self.PACKET_SIZE)
        ]

    def decode(self, func=show_cue):
        """
        Stream.decode reads self.tsdata to find SCTE35 packets.
        func can be set to a custom function that accepts
        a fu.Cue instance as it's only argument.
        """
        if not self._find_start():
            return
        for pkt in self.iter_pkts():
            cue = self._parse(pkt)
            if cue:
                func(cue)
        return False

    def decode_fu(self, func=show_cue):
        """
        Stream.decode_fu decodes
        num_pkts packets at a time.
        Super Fast with pypy3.
        """
        num_pkts = 3000
        for chunk in self.iter_pkts(num_pkts=num_pkts):
            _ = [func(cue) for cue in self._mk_pkts(chunk) if cue]
            del _
        return False

    def decode_next(self):
        """
        Stream.decode_next returns the next
        SCTE35 cue as a fu.Cue instance.
        """
        for pkt in self.iter_pkts():
            cue = self._parse(pkt)
            if cue:
                return cue
        return False

    def decode_program(self, the_program, func=show_cue):
        """
        Stream.decode_program limits SCTE35 parsing
        to a specific MPEGTS program.
        """
        self.the_program = the_program
        return self.decode(func)

    def decode_pids(self, scte35_pids=[], func=show_cue):
        """
        Stream.decode_pids takes a list of SCTE-35 Pids parse
        and an optional call back function to run when a Cue is found.
        if scte35_pids is not set, all scte35 pids will be parsed.
        """
        self.the_scte35_pids = scte35_pids
        return self.decode(func)

    def decode_start_time(self):
        """
        decode_start_time
        """
        self.decode(func=no_op)
        if len(self.start.values()) > 0:
            return self.start.popitem()[1]
        return False

    def proxy(self, func=show_cue_stderr):
        """
        Stream.decode_proxy writes all ts packets are written to stdout
        for piping into another program like mplayer.
        SCTE-35 cues are print2`ed to stderr.
        """
        for pkt in self.iter_pkts():
            cue = self._parse(pkt)
            if cue:
                func(cue)
            sys.stdout.buffer.write(pkt)
        return False

    def show(self):
        """
        displays streams that will be
        parsed for SCTE-35.
        """
        self.info = True
        for pkt in self.iter_pkts():
            self._parse_info(pkt)
        if self.maps.prgm.keys():
            sopro = sorted(self.maps.prgm.items())
            for k, vee in sopro:
                if len(vee.streams.items()) > 0:
                    print2(f"\nProgram: {k}")
                    vee.show()
            return True

    def show_pts(self):
        """
        show_pts displays current pts by pid.
        """
        print("\tPrgm\tPTS")
        for pkt in self.iter_pkts():
            pid = self._parse_info(pkt)
            if pid in self.pids.pcr:
                if self._pusi_flag(pkt) and pid != 0:
                    self._parse_pts(pkt, pid)
                    print(f"\t{self.pid2prgm(pid)}\t{self.pid2pts(pid)}")

    def pts(self):
        """
        pts returns a dict of  program:pts
        """
        return self.maps.prgm_pts

    def pid2prgm(self, pid):
        """
        pid2prgm takes a pid,
        returns the program
        """
        prgm = 1
        if pid in self.maps.pid_prgm:
            prgm = self.maps.pid_prgm[pid]
        return prgm

    def pid2pts(self, pid):
        """
        pid2pts takes a pid
        returns the current pts
        """
        prgm = self.pid2prgm(pid)
        if prgm not in self.maps.prgm_pts:
            return False
        return self.as_90k(self.maps.prgm_pts[prgm])

    def pid2pcr(self, pid):
        """
        pid2pcr takes a pid
        returns the current pcr
        """
        prgm = self.pid2prgm(pid)
        if prgm not in self.maps.prgm_pcr:
            return False
        return self.as_90k(self.maps.prgm_pcr[prgm])

    def _unpad_afc(self, pkt):
        if self._afc_flag(pkt[3]):
            pkt = pkt[:4] + self._unpad(pkt[4:])
        return pkt

    def _unpad(self, bites):
        pad = 255
        one = 1
        if not bites:
            return b""
        if bites[0] in [pad]:
            self._unpad(bites[one:])
        return bites

    def _mk_packet_data(self, pid):
        prgm = self.maps.pid_prgm[pid]
        pdata = PacketData(pid, prgm)
        pdata.mk_pcr(self.maps.prgm_pcr)
        pdata.mk_pts(self.maps.prgm_pts)
        return pdata

    def _parse_cc(self, pkt, pid):
        last_cc = None
        c_c = pkt[3] & 0xF
        if pid in self.maps.pid_cc:
            last_cc = self.maps.pid_cc[pid]
            good = (last_cc, ((last_cc + 1) % 16))
            if c_c not in good:
                print2(f" # BAD --> pid:\t{hex(pid)}\tlast cc:\t{last_cc}\tcc:\t{c_c}")
        self.maps.pid_cc[pid] = c_c

    def _parse_pts(self, pkt, pid):
        """
        parse pts and store by program key
        in the dict Stream._pid_pts
        """
        payload = self._parse_payload(pkt)
        if len(payload) > 13:
            if self._pts_flag(payload):
                pts = (payload[9] & 14) << 29
                pts |= payload[10] << 22
                pts |= (payload[11] >> 1) << 15
                pts |= payload[12] << 7
                pts |= payload[13] >> 1
                prgm = self.pid2prgm(pid)
                self.maps.prgm_pts[prgm] = pts
                if prgm not in self.start:
                    self.start[prgm] = pts

    def _parse_pcr(self, pkt, pid):
        if self._afc_flag(pkt[3]):
            pcr = pkt[6] << 25
            pcr |= pkt[7] << 17
            pcr |= pkt[8] << 9
            pcr |= pkt[9] << 1
            pcr |= pkt[10] >> 7
            prgm = self.pid2prgm(pid)
            self.maps.prgm_pcr[prgm] = pcr

    def _parse_payload(self, pkt):
        """
        _parse_payload returns the packet payload
        """
        head_size = 4
        if self._afc_flag(pkt[3]):
            pkt = self._unpad_afc(pkt)
            afl = pkt[4]
            head_size += afl + 1  # +1 for afl byte
        return pkt[head_size:]

    def _parse_tables(self, pkt, pid):
        """
        _parse_tables parse for
        PAT, PMT,  and SDT tables
        based on pid of the pkt
        """
        pay = self._parse_payload(pkt)
        if self._same_as_last(pay, pid):
            return False
        if pid in self.pids.pmt:
            return self._parse_pmt(pay, pid)
        if pid == self.pids.PAT_PID:
            return self._parse_pat(pay)
        if pid == self.pids.SDT_PID and self.info:
            return self._parse_sdt(pay)
        return False

    def _parse_info(self, pkt):
        """
        _parse_info parses the packet for tables
        and returns the pid
        """
        pid = self._parse_pid(pkt[1], pkt[2])
        if pid in self.pids.tables:
            self._parse_tables(pkt, pid)
        return pid

    def _parse(self, pkt):
        cue = False
        pid = self._parse_info(pkt)
        # self._parse_cc(pkt, pid)
        if self._pcr_flag(pkt):
            self._parse_pcr(pkt, pid)
        if self._pusi_flag(pkt):
            self._parse_pts(pkt, pid)
        if self._pid_has_scte35(pid):
            cue = self._parse_scte35(pkt, pid)
        return cue

    def _pid_has_scte35(self, pid):
        if pid in self.pids.scte35 or pid in self.pids.maybe_scte35:
            return True
        return False

    def _chk_partial(self, pay, pid, sep):
        if pid in self.maps.partial:
            pay = self.maps.partial.pop(pid) + pay
        return self._split_by_idx(pay, sep)

    def _same_as_last(self, pay, pid):
        same = False
        if pid in self.maps.last:
            same = (False, True)[pay == self.maps.last[pid]]
        if not same:
            self.maps.last[pid] = pay
        return same

    def _section_incomplete(self, pay, pid, seclen):
        # + 3 for the bytes before section starts
        if (seclen + 3) > len(pay):
            self.maps.partial[pid] = pay
            return True
        return False

    def _parse_cue(self, pay, pid):
        packet_data = None
        packet_data = self._mk_packet_data(pid)
        cue = Cue(pay, packet_data)
        if cue.decode():
            return cue
        return False

    def _strip_scte35_pes(self, pkt):
        pay = self._parse_payload(pkt)
        if self.SCTE35_PES_START in pay:
            # print2(f"# Stripping PES Header from SCTE35 @ {self.pid2pts(pid)}")
            pay = pay.split(self.SCTE35_PES_START, 1)[-1]
            peslen = pay[4] + 5  # PES header length
            pay = pay[peslen:]
        return pay

    def _mk_scte35_payload(self, pkt, pid):
        pay = self._strip_scte35_pes(pkt)
        if not pay:
            return False
        pay = self._chk_partial(pay, pid, self.SCTE35_TID)
        if not pay:
            if pid in self.pids.maybe_scte35:
                self.pids.maybe_scte35.remove(pid)
                return False
        if pay[13] == self.show_null:
            return False
        return pay

    def _parse_scte35(self, pkt, pid):
        """
        parse a scte35 cue from one or more packets
        """
        pay = self._mk_scte35_payload(pkt, pid)
        if not pay:
            return False
        seclen = self._parse_length(pay[1], pay[2])
        if self._section_incomplete(pay, pid, seclen):
            return False
        pay = pay[: seclen + 3]
        cue = self._parse_cue(pay, pid)
        return cue

    def _parse_sdt(self, pay):
        """
        _parse_sdt parses the SDT for program metadata
        """
        pay = self._chk_partial(pay, self.pids.SDT_PID, self.SDT_TID)
        if not pay:
            return False
        seclen = self._parse_length(pay[1], pay[2])
        if self._section_incomplete(pay, self.pids.SDT_PID, seclen):
            return False
        idx = 11
        while idx < seclen + 3:
            service_id = self._parse_program(pay[idx], pay[idx + 1])
            idx += 3
            dloop_len = self._parse_length(pay[idx], pay[idx + 1])
            idx += 2
            i = 0
            while i < dloop_len:
                if pay[idx] == 0x48:
                    i += 3
                    spnl = pay[idx + i]
                    i += 1
                    service_provider_name = pay[idx + i : idx + i + spnl]
                    i += spnl
                    snl = pay[idx + i]
                    i += 1
                    service_name = pay[idx + i : idx + i + snl]
                    i += snl
                    if service_id not in self.maps.prgm:
                        self.maps.prgm[service_id] = ProgramInfo()
                    pinfo = self.maps.prgm[service_id]
                    pinfo.provider = service_provider_name
                    pinfo.service = service_name
                i = dloop_len
                idx += i
                return True
        return False

    def _parse_pat(self, pay):
        """
        parse program association table
        for program to pmt_pid mappings.
        """
        pay = self._chk_partial(pay, self.pids.PAT_PID, b"")
        seclen = self._parse_length(pay[2], pay[3])
        if self._section_incomplete(pay, self.pids.PAT_PID, seclen):
            return False
        seclen -= 5  # pay bytes 4,5,6,7,8
        idx = 9
        chunk_size = 4
        while seclen > 4:  #  4 bytes for crc
            program_number = self._parse_program(pay[idx], pay[idx + 1])
            if program_number > 0:
                pmt_pid = self._parse_pid(pay[idx + 2], pay[idx + 3])
                self.pids.pmt.add(pmt_pid)
                self.pids.tables.add(pmt_pid)
            seclen -= chunk_size
            idx += chunk_size
        return True

    def _mk_pmt_payload(self, pay, pid):
        seclen = False
        program_number = False
        pay = self._chk_partial(pay, pid, self.PMT_TID)
        if not pay:
            return False
        seclen = self._parse_length(pay[1], pay[2])
        if self._section_incomplete(pay, pid, seclen):
            return False
        program_number = self._parse_program(pay[3], pay[4])
        if self.the_program and (program_number != self.the_program):
            return False
        return pay, seclen, program_number

    def _parse_pmt(self, pay, pid):
        """
        parse program maps for streams
        """
        pay, seclen, program_number = self._mk_pmt_payload(pay, pid)
        if not program_number:
            return False
        pcr_pid = self._parse_pid(pay[8], pay[9])
        if program_number not in self.maps.prgm:
            self.maps.prgm[program_number] = ProgramInfo()
        pinfo = self.maps.prgm[program_number]
        pinfo.pid = pid
        pinfo.pcr_pid = pcr_pid
        self.pids.pcr.add(pcr_pid)
        self.maps.pid_prgm[pcr_pid] = program_number
        proginfolen = self._parse_length(pay[10], pay[11])
        idx = 12 + proginfolen
        si_len = seclen - (9 + proginfolen)
        self._parse_program_streams(si_len, pay, idx, program_number)
        return True

    def _parse_program_streams(self, si_len, pay, idx, program_number):
        """
        parse the elementary streams
        from a program
        """
        # 5 bytes for stream_type info
        chunk_size = 5
        end_idx = (idx + si_len) - chunk_size
        while idx < end_idx:
            stream_type, pid, ei_len = self._parse_stream_type(pay, idx)
            pinfo = self.maps.prgm[program_number]
            pinfo.streams[pid] = stream_type
            idx += chunk_size + ei_len
            self.maps.pid_prgm[pid] = program_number

    def _parse_stream_type(self, pay, idx):
        """
        extract stream pid and type
        """
        stream_type = hex(pay[idx])
        el_pid = self._parse_pid(pay[idx + 1], pay[idx + 2])
        ei_len = self._parse_length(pay[idx + 3], pay[idx + 4])
        self._set_scte35_pids(el_pid, stream_type)
        return stream_type, el_pid, ei_len

    def _set_scte35_pids(self, pid, stream_type):
        """
        if stream_type is 0x06 or 0x86
        add it to self._scte35_pids.
        """
        if stream_type in ["0x86"]:
            self.pids.scte35.add(pid)
        if stream_type in ["0x06", "0x6"]:
            self.pids.maybe_scte35.add(pid)



#!/usr/bin/env python3
"""
fu.hls  hls.py
(replaces showcues)

"""

import datetime
import sys
import time
from collections import deque
from fu import Segment, Cue
from m3ufu import M3uFu, TagParser, HEADER_TAGS
from new_reader import reader

REV = "\033[7m"
NORM = "\033[27m"
# REV = ""
# NORM = ""
SUB = "\t"
NSUB = f"\n{SUB}"
ROLLOVER = 95443.717678
HEADER_TAGS = list(HEADER_TAGS)
HEADER_TAGS.append("#EXTM3U")


def atoif(value):
    """
    atoif converts ascii to (int|float)
    """
    if isinstance(value, (int, float)):
        return value
    if isinstance(value, str):
        if "." in value:
            try:
                value = float(value)
                return value
            except:
                pass
        else:
            try:
                value = int(value)
                return value
            except:
                pass
    return None


def iso8601():
    """
    return UTC time in iso8601 format.

    '2023-05-11T15:55:51.'

    """
    return f"{datetime.datetime.utcnow().isoformat()[:-4]}Z "


class Scte35Profile:
    """
    Scte35Profile Manages the Scte35 Parsing Profile
    """

    def __init__(self):
        """
        A Scte35Profile
        """
        self.expand_cues = False  # Show SCTE-35 Cues fully expanded.
        self.parse_segments = True  # Parse Segments for SCTE-35.
        self.parse_manifests = True  # Parse m3u8 files for SCTE-35 HLS tags.
        self.hls_tags = [
            "#EXT-OATCLS-SCTE35",
            "#EXT-X-CUE-OUT-CONT",
            "#EXT-X-DATERANGE",
            "#EXT-X-SCTE35",
            "#EXT-X-CUE-IN",
            "#EXT-X-CUE-OUT",
        ]  #  Parse these types of HLS SCTE-35 tags.
        self.command_types = [6, 5]  # Which SCTE-35 Commands to parse.
        self.descriptor_tags = [
            2,
        ]  # Which Descriptors to parse.
        # Which Descriptor Segmentation Types IDs should be parsed
        self.starts = [0x22, 0x30, 0x32, 0x34, 0x36, 0x44, 0x46]
        self.seg_type = 0x23

    # self.stops = [0x23, 0x31, 0x33, 0x35, 0x37, 0x45, 0x47]

    def write_profile(self, pro_file):
        """
        write_profile writes sc.profile for editing.
        """
        with open(pro_file, "w") as pro_f:
            for que, vee in vars(self).items():
                line = f"{que} = "
                if isinstance(vee, list):
                    for item in vee:
                        if isinstance(item, int):
                            line = f"{line}{hex(item)},"
                        else:
                            line = f"{line}{item},"
                if isinstance(vee, bool):
                    line = f"{line}{vee}"
                if line.endswith(","):
                    line = line[:-1]
                pro_f.write(line + "\n")

    def show_profile(self, headline):
        """
        show_profile displays profile settings.
        """
        print(f"\n\n\t{REV}{headline}{NORM}\n")
        for que, vee in vars(self).items():
            if isinstance(vee, list):
                if isinstance(vee[0], int):
                    vee = [hex(eye) for eye in vee]
            print(f"\t{que} = {vee}\n")
            time.sleep(0.3)

    def clean_n_split(self, line):
        """
        clean_n_split a line.
        """
        this, that = None, None
        bad = [" ", "\n", "\t", '"', "'"]
        for bee in bad:
            line = line.replace(bee, "")
        if line:
            this, that = line.split("=", 1)
            that = list(that.split(","))
        if line.startswith("#") or line.startswith("//"):
            this = None
        return this, that

    def format4profile(self, this, that):
        """
        format4profile formats data read from hls.profile for internal use.
        """
        if this is None or that is None:
            return
        this = this.lower()
        if this.startswith("parse") or this.startswith("expand"):
            if that[0].lower().startswith("f"):
                that = False
            else:
                that = True
        if this in ["command_types", "descriptor_tags", "starts"]:
            new_that = []
            for s in that:
                if s.lower().startswith("0x"):
                    new_that.append(int(s, 16))
                else:
                    new_that.append(int(s))
            that = new_that
        self.__dict__.update({this: that})

    def read_profile(self, pro_file):
        """
        read_profile reads sc.profile
        """
        try:
            with open(pro_file, "r") as pro_file:
                for line in pro_file:
                    this, that = self.clean_n_split(line)
                    self.format4profile(this, that)
        except:
            pass
        finally:
            self.show_profile("Profile:")

    def set_pts(self, cue):
        """
        set_pts cue.command.pts_time +
        cue.info_section.pts_adjustment
        % ROLLOVER
        """
        pts = cue.command.pts_time + cue.info_section.pts_adjustment
        return pts % ROLLOVER

    def validate_cue(self, cue):
        """
        validate_cue use the parsing profile to validate a SCTE-35 Cue.
        """
        pts = None
        line = None
        cue.decode()
        if cue.command.command_type in self.command_types:
            if "pts_time" in vars(cue.command) and cue.command.pts_time:
                pts = self.set_pts(cue)
            if self.expand_cues:
                cue.show()
            if cue.command.command_type == 5:
                line = self.validate_splice_insert(cue)
        if cue.command.command_type == 6:
            line = self.validate_time_signal(cue)
        return pts, line

    def validate_splice_insert(self, cue):
        """
        validate_splice_insert is named appropriately.
        """
        line = None
        if cue.command.out_of_network_indicator:
            if "break_duration" in vars(cue.command) and cue.command.break_duration:
                duration = cue.command.break_duration
                line = f"#EXT-X-CUE-OUT:{duration}\n"
                return line
        else:
            line = "#EXT-X-CUE-IN\n"
            return line
        return line

    def validate_time_signal(self, cue):
        """
        validate_time_signal is named appropriately.
        """
        line = None
        for dscptr in cue.descriptors:
            if dscptr.tag in self.descriptor_tags:
                if dscptr.segmentation_type_id in self.starts:
                    self.seg_type = dscptr.segmentation_type_id + 1
                    if "segmentation_duration" in vars(dscptr):
                        duration = dscptr.segmentation_duration
                        line = f"#EXT-X-CUE-OUT:{duration}\n"
                        return line
                if dscptr.segmentation_type_id == self.seg_type:
                    line = "#EXT-X-CUE-IN\n"
                    self.seg_type = None
                    return line
        return line


class Pane:
    """
    Pane class. Sliding_Window slides Panes
    """

    def __init__(self, media, lines):
        self.media = media
        self.lines = lines

    def get(self):
        """
        get merges self.lines and self.media for
        writing m3u8 files.
        """
        all_lines = self.lines + [self.media]
        return "".join(all_lines)


class SlidingWindow:
    """
    The Sliding Window class
    """

    def __init__(self, size=101):
        self.size = size
        self.panes = deque()
        self.delete = False

    def pop_pane(self):
        """
        pop_pane removes the first item in self.panes
        """
        if len(self.panes) > self.size:
            self.panes.popleft()

    def all_panes(self):
        """
        all_panes returns the current window panes joined.
        """
        return "\n".join(set([a_pane.get() for a_pane in self.panes]))

    def slide_panes(self, a_pane):
        """
        slide calls self.push_pane with a_pane
        and then calls self.pop_pane to trim self.panes
        as needed.
        """
        self.panes.append(a_pane)
        self.pop_pane()


class AacParser:
    """
    AacParser parses aac segments.
    """

    applehead = b"com.apple.streaming.transportStreamTimestamp"

    @staticmethod
    def is_header(header):
        """
        is_header tests aac and ac3 files for ID3 headers.
        """
        if header[:3] == b"ID3":
            return True
        return False

    @staticmethod
    def id3_len(header):
        """
        id3_len parses the length value from ID3 headers
        """
        id3len = int.from_bytes(header[6:], byteorder="big")
        return id3len

    @staticmethod
    def syncsafe5(somebytes):
        """
        syncsafe5 parses PTS from ID3 tags.
        """
        lsb = len(somebytes) - 1
        syncd = 0
        for idx, bite in enumerate(somebytes):
            syncd += bite << ((lsb - idx) << 3)
        return round(syncd / 90000.0, 6)

    def parse(self, media):
        """
        aac_pts parses the ID3 header tags in aac and ac3 audio files
        """
        pts = 0
        aac = reader(media)
        header = aac.read(10)
        if self.is_header(header):
            id3len = self.id3_len(header)
            data = aac.read(id3len)
            if self.applehead in data:
                try:
                    pts = float(data.split(self.applehead)[1].split(b"\x00", 2)[1])
                except:
                    pts = self.syncsafe5(data.split(self.applehead)[1][:9])
        return round((pts % ROLLOVER), 6)


class CuePuller:
    """
    CuePuller is the main object.
    """

    def __init__(self):
        self.media = deque()
        self.sidecar = "hls.sidecar"
        self.dumpfile = "hls.dump"
        self.flat = "hlsflat.m3u8"
        self.pro_file = "hls.profile"
        self.m3u8 = "hls.m3u8"
        self.last_dump_line = None
        self.base_uri = None
        self.iv = None
        self.key_uri = None
        self.last_iv = None
        self.last_key_uri = None
        self.break_timer = None
        self.break_duration = None
        self.reload = True
        self.sleep_duration = 0
        self.window_size = None
        self.sliding_window = SlidingWindow()
        self.cue_state = None
        self.last_cue = None
        self.headers = []
        self.pts = 0
        self.cont_resume = False
        self.first_segment = True
        self.hls_pts = "HLS"
        self.prof = Scte35Profile()
        self.prof.read_profile(self.pro_file)
        self.clear_files()

    @staticmethod
    def clear():
        """
        clear previous line.
        """
        print(" " * 80, end="\r")
        print(" " * 80, file=sys.stderr, end="\r", flush=True)

    def clear_files(self):
        """
        clear_files clobbers the appended files
        self.sidecar, self.dumpfile, self.flat, self.m3u8
        when showcues is started.
        """
        for sidef in [self.sidecar, self.dumpfile, self.flat, self.m3u8]:
            with open(sidef, "w+") as side_file:  # touch
                pass

    def chk_aes(self, line):
        """
        chk_aes checks for AES encryption
        """
        if "#EXT-X-KEY" in line:
            tags = TagParser([line]).tags
            if "URI" in tags["#EXT-X-KEY"]:
                self.key_uri = tags["#EXT-X-KEY"]["URI"]
                if not self.key_uri.startswith("http"):
                    re_uri = self.base_uri + self.key_uri
                    self.key_uri = re_uri
                if "IV" in tags["#EXT-X-KEY"]:
                    self.iv = tags["#EXT-X-KEY"]["IV"]

    def to_sidecar(self, pts, line):
        """
        to_sidecar writes (pts,hls tag) pairs to the sidecar file.
        """
        with open(self.sidecar, "a") as sidecar:
            sidecar.write(f"{round(pts,6)},{line}\n")

    def to_dump(self, pts, line):
        """
        to_dump copies all SCTE-35 lines to self.dumpfile.
        """
        with open(self.dumpfile, "a") as dump:
            dump_line = f"{pts},{line}\n"
            if dump_line != self.last_dump_line:
                dump.write(dump_line)
                self.last_dump_line = dump_line

    def media_stuff(self):
        """
        media_stuff trims segment URI to just the file name.
        """
        media = self.media[-1]
        short_media = media.rsplit("/", 1)[1].split("?", 1)[0]
        return f"Media: {short_media.strip()}"

    def cue_stuff(self):
        """
        cue_stuff returns self.cue_state formated.
        """
        return f"Cue {REV}{self.cue_state} {NORM}"

    def diff_stuff(self):
        """
        diff gonzo returns formated self.break_timer
        and if possible the difference between the actual break duration
        and the specified SCTE-35 break duration.
        """
        if self.break_timer is not None:
            if not self.break_duration:
                return f"{NSUB}Break Timer: {round(self.break_timer,6)}"
            return (NSUB).join(
                [
                    f"{NSUB}Timer: { round(self.break_timer,6)}",
                    f"Duration: {self.break_duration}",
                    f"Diff: {round(self.break_timer - self.break_duration,6)}",
                ]
            )
        return ""

    def dur_stuff(self):
        """
        dur_stuff returns self.break_duration formated.
        """
        return f"{NSUB}Duration: {self.break_duration}"

    def pts_stuff(self):
        """
        pts_stuff returns  PTS formated.
        """
        return f"{NSUB}{self.hls_pts}: {self.pts}"

    def set_cue_state(self, cue, line):
        """
        set_cue_state determines cue_state

        """
        if cue.encode() == self.last_cue:
            return ""
        self.last_cue = cue.encode()
        if "CONT" not in line:
            head = f"\n{iso8601()}{REV}{line}{NORM}{self.pts_stuff()} {REV} Splice Point {NORM}"
            if line.startswith("#EXT-X-CUE-IN") and self.cue_state == "CONT":
                self.cue_state = "IN"
                self.to_sidecar(self.pts, line)
                self.clear()
                print(f"{head}{self.diff_stuff()}{NSUB}{self.media_stuff()}\n")
                self.reset_break()
                return line
            if line.startswith("#EXT-X-CUE-OUT") and self.cue_state in [None, "IN"]:
                self.reset_break()
                self.cue_state = "OUT"
                self.break_timer = 0.0
                if ":" in line:
                    self.break_duration = atoif(line.split(":")[1])
                self.to_sidecar(self.pts, line)
                self.clear()
                print(f"{head}{self.dur_stuff()}{NSUB}{self.media_stuff()}\n")
                return line
        if "CONT" in line and self.cue_state in ["OUT", "CONT"]:
            self.to_sidecar(self.pts, line)
            self.cue_state = "CONT"
            return line
        return line

    def invalid(self, line):
        """
        invalid print invalid SCTE-35 HLS tags
        """
        self.clear()
        print(
            f"\n{iso8601()}{REV} Skipped {NORM}  {line}{self.pts_stuff()}{NSUB}{self.media_stuff()}\n"
        )
        return "## " + line

    def show_tags(self, tags):
        try:
            for que, vee in tags.items():
                print(f"{SUB}{que}: {vee}")
        except:
            return

    def _set_break_timer(self, line, cont_tags):
        """
        _set_break_timer sets self.break_timer to ElaspsedTime
        read from a CUE-OUT-CONT tag or to 0.0.
        """
        if self.break_timer:
            return
        if "ElapsedTime" in cont_tags:
            self.break_timer = cont_tags["ElapsedTime"]
        else:
            try:
                self.break_timer = round(float(line.split(":", 1)[1].split("/")[0]), 3)
            except:
                self.break_timer = 0.0
        print(f"{iso8601()}{REV} Break Timer {NORM}to {self.break_timer}\n")
        time.sleep(0.1)

    def _set_break_duration(self, line, cont_tags):
        """
        __set_break_duration sets self.break_duration from
        a CUE-OUT-CONT tag or from a CUE-OUT tag.
        """
        if self.break_duration:
            return
        if "Duration" in cont_tags:
            self.break_duration = cont_tags["Duration"]
        else:
            try:
                self.break_duration = round(
                    float(line.split(":", 1)[1].split("/")[1]), 3
                )
            except:
                self.break_duration = None
        if self.break_duration:
            print(f"{iso8601()}{REV}Break Duration {NORM}to {self.break_duration}\n")
            time.sleep(0.1)

    def chk_x_cue_out_cont(self, tags, line):
        """
        chk_x_cue_out_const processes
        #EXT-X-CUE-OUT-CONT tags
        """
        cont_tags = tags["#EXT-X-CUE-OUT-CONT"]
        if self.cue_state not in ["OUT", "CONT"] and not self.first_segment:
            return None
        if self.first_segment:
            print(f"{iso8601()}{REV} Resuming Ad Break {NORM}\n")
            self.cue_state = "CONT"
            self._set_break_timer(line, cont_tags)
            self._set_break_duration(line, cont_tags)
        return self.auto_cont()

    def chk_x_cue_in(self, tags, line):
        """
        chk_x_cue_in processes
        #EXT-X-CUE-IN tags.
        """
        return self.set_cue_state(line, line)

    def chk_x_cue_out(self, tags, line):
        """
        chk_x_cue_out processes
        #EXT-X-CUE-OUT tags
        """
        return self.set_cue_state(line, line)

    def chk_x_scte35(self, tags, line):
        """
        chk_x_scte35 handles #EXT-X-SCTE35 tags.
        """
        if "CUE" in tags["#EXT-X-SCTE35"]:
            cue = Cue(tags["#EXT-X-SCTE35"]["CUE"])
            pts, new_line = self.prof.validate_cue(cue)
            if pts and new_line:
                return self.set_cue_state(tags["#EXT-X-SCTE35"]["CUE"], new_line)
        return self.invalid(line)

    def chk_x_daterange(self, tags, line):
        """
        chk_x_daterange handles #EXT-X-DATERANGE tags.
        """
        self.show_tags(tags["#EXT-X-DATERANGE"])
        for scte35_tag in ["SCTE35-OUT", "SCTE35-IN"]:
            if scte35_tag in tags["#EXT-X-DATERANGE"]:
                cue = Cue(tags["#EXT-X-DATERANGE"][scte35_tag])
                pts, new_line = self.prof.validate_cue(cue)
                if pts and new_line:
                    return self.set_cue_state(
                        tags["#EXT-X-DATERANGE"][scte35_tag], new_line
                    )
        return self.invalid(line)

    def chk_x_oatcls(self, tags, line):
        """
        chk_x_oatcls handles
        #EXT-OATCLS-SCTE35
        HLS tags.
        """
        cue = Cue(tags["#EXT-OATCLS-SCTE35"])
        pts, new_line = self.prof.validate_cue(cue)
        if pts and new_line:
            if abs(pts - self.pts) > 5:  # Handle Cues out of sync with video PTS
                pts = self.pts
            return self.set_cue_state(tags["#EXT-OATCLS-SCTE35"], new_line)
        return self.invalid(line)

    def scte35(self, line):
        """
        scte35 processes SCTE-35 related tags.
        """
        scte35_map = {
            "#EXT-X-DATERANGE": self.chk_x_daterange,
            "#EXT-X-SCTE35": self.chk_x_scte35,
            "#EXT-X-CUE-OUT-CONT": self.chk_x_cue_out_cont,
            "#EXT-OATCLS-SCTE35": self.chk_x_oatcls,
            "#EXT-X-CUE-IN": self.chk_x_cue_in,
            "#EXT-X-CUE-OUT": self.chk_x_cue_out,
        }
        tags = TagParser([line]).tags
        for key in scte35_map.keys():
            if key in line:
                self.to_dump(self.pts, line)
        if self.prof.parse_manifests:
            for que, vee in scte35_map.items():
                if que in line:
                    if que not in self.prof.hls_tags:
                        return None
                    return vee(tags, line)
        return line

    def auto_cont(self):
        self.cue_state = "CONT"
        line = (
            f"#EXT-X-CUE-OUT-CONT:{round(self.break_timer,3)}/{self.break_duration}\n"
        )
        line = self.set_cue_state(line, line)
        return line

    def auto_cuein(self, line):
        """
        auto_cuein handles cue.command.auto-return
        """
        if self.cue_state == "CONT":
            if self.break_timer and self.break_duration:
                if self.break_timer >= self.break_duration:
                    self.cue_state = "IN"
                    self.clear()
                    print(
                        f"{iso8601()}{REV} AUTO CUE-IN {NORM}{self.pts_stuff()}{self.diff_stuff()}{NSUB}{self.media_stuff()}\n"
                    )
                    self.reset_break()
                    self.to_sidecar(self.pts, "#AUTO\n#EXT-X-CUE-IN\n")
                    return "#AUTO\n#EXT-X-CUE-IN\n" + line
        return line

    def reset_break(self):
        """
        reset_break resets
        break_duration, break_timer,
        and cue_state after a CUE-IN
        """
        if self.cue_state == "IN":
            self.break_duration = None
            self.break_timer = None
            self.cue_state = None

    def extinf(self, line):
        """
        extinf parses lines that start with #EXTINF
        for the segment duration.
        """
        tags = TagParser([line]).tags
        if "#EXTINF" in tags:
            if isinstance(tags["#EXTINF"], str):
                tags["#EXTINF"] = tags["#EXTINF"].rsplit(",", 1)[0]
            seg_time = round(atoif(tags["#EXTINF"]), 6)
            line = self.auto_cuein(line)
            if self.pts is not None:
                self.pts += seg_time
            if self.break_timer is not None:
                self.break_timer += seg_time
        return line

    def print_time(self):
        """
        print_time prints wall clock and pts.
        """
        gonzo = ""
        if self.break_timer:
            gonzo = f"{REV} Break {NORM} {round(self.break_timer,3)}"
            if self.break_duration:
                gonzo = f"{gonzo} / {round(self.break_duration,3)}"
        print(
            f"\r\r{iso8601()}{REV} {self.hls_pts} {NORM} {self.pts} {gonzo}",
            end="\r",
            file=sys.stderr,
            flush=True,
        )

    def ts_pts(self, seg):
        """
        ts_pts set pts from segment
        """
        if seg.pts_start:
            self.pts = seg.pts_start
            self.print_time()
            self.hls_pts = "PTS"

    def ts_cues(self, seg):
        """
        ts_cues process SCTE-35 cues
        found in a segment.
        """
        for cue in seg.cues:
            if "pts" in vars(cue.packet_data) and cue.packet_data.pts:
                self.pts = cue.packet_data.pts
            if cue.encode() != self.last_cue:
                self.last_cue = cue.encode()
                self.ts_set_cue(cue)

    def ts_set_cue(self, cue):
        """
        ts_set_cue validate and set cue from ts segment.
        """
        cue_pts, line = self.prof.validate_cue(cue)
        if cue_pts and line:
            self.set_cue_state(cue.encode(), line)
            self.clear()
            print(
                (NSUB).join(
                    [
                        f"\n{iso8601()}{REV} SCTE-35 {NORM}",
                        f"Stream PTS: {round(self.pts,6)}",
                        f"PreRoll: {round(cue_pts - self.pts,6)}",
                        f"Splice Point: {round(cue_pts,6)}",
                        f"Type: {cue.command.name}",
                        f"{self.media_stuff()}\n",
                    ]
                )
            )

    def chk_ts(self, this):
        """
        chk_ts  check MPEGTS for PTS and SCTE-35.
        """
        if ".ts" in this:
            if self.first_segment:
                Segment(this, key_uri=self.key_uri, iv=self.iv).show()
            seg = Segment(this, key_uri=self.key_uri, iv=self.iv)
            seg.shushed()
            seg.decode()
            self.ts_pts(seg)
            if self.prof.parse_segments:
                self.ts_cues(seg)
            self.print_time()

    def chk_aac(self, this):
        """
        chk_aac check aac and ac3  HLS audio segments
        for PTS in ID3 header tags.
        """
        if ".aac" in this or ".ac3" in this:
            aac_parser = AacParser()
            pts = aac_parser.parse(this)
            if pts:
                self.pts = pts
                self.hls_pts = "PTS"
                self.print_time()

    def new_media(self, this):
        """
        new_media check to see
        if the media is new in a
        live sliding window
        """
        if this not in self.media:
            self.media.append(this)
            if len(self.media) > self.window_size + 1:
                self.media.popleft()
            return True
        return False

    def parse_target_duration(self, line):
        """
        parse_target_duration reads target duration
        off the manifest to set self.sleep_duration.
        self.sleep_duration is used to throttle manifest
        requests.
        """
        if "TARGETDURATION" in line:
            if self.sleep_duration == 0:
                target_duration = atoif(line.split(":")[1])
                self.sleep_duration = round(target_duration * 0.5, 3)
                print(f"{SUB}{REV} Target Duration {NORM} {target_duration}\n")

    def mk_window_size(self, lines):
        """
        mk_window_size sets the sliding window size
        for the output to match that off the input and
        determine how long to keep media data info
        for segments.
        """
        if not self.window_size:
            self.window_size = len([line for line in lines if "#EXTINF:" in line])
            self.sliding_window.size = self.window_size
            print(f"{SUB}{REV} Window Size {NORM} {self.window_size}\n")

    def update_cue_state(self):
        """
        update_cue_state changes CUE state.
        """
        if self.cue_state == "OUT":
            self.cue_state = "CONT"
        if self.cue_state == "IN":
            self.cue_state = None

    @staticmethod
    def decode_lines(lines):
        """
        decode_lines convert bytes to ascii
        """
        return [line.decode() for line in lines]

    def parse_line(self, line):
        if "#EXT-X-PROGRAM-DATE-TIME" in line:
            return None

        if "#EXTINF:" in line:
            line = self.extinf(line)
            return line
        self.chk_aes(line)
        line = self.scte35(line)
        return line

    def parse_header(self, line):
        """
        parse_headers parses m3u8 files for HLS header tags.
        """
        if "#EXT-X-PROGRAM-DATE-TIME" in line:
            return False
        splitline = line.split(":", 1)
        if splitline[0] in HEADER_TAGS:
            self.parse_target_duration(line)
            self.headers.append(line)
            return True
        return False

    def chk_endlist(self, line):
        """
        chk_endlist disables manifest reloading
        if line contains ENDLIST tag.
        """
        if "#EXT-X-ENDLIST" in line:
            self.reload = False

    def write_flat(self, lines, media):
        """
        write_flat flatten out the sliding window
        and write all data to flat.m3u8.
        """
        with open(self.flat, "a") as flat:
            if self.first_segment:
                flat.write("#EXTM3U\n")
                for header in self.headers:
                    flat.write(header)
            for line in lines:
                flat.write(line)
            flat.write(media)

    def write_manifest(self):
        """
        write_manifest write data to sc.m3u8
        with profile rules applied.
        """
        with open(self.m3u8, "w") as out:
            out.write("#EXTM3U\n")
            out.write("".join(self.headers))
            out.write(self.sliding_window.all_panes())

    def _parse_new_media(self, lines, media):
        self.write_flat(lines, media)
        parsed = [self.parse_line(line) for line in lines]
        lines = [line for line in parsed if line is not None]
        media = media.replace("\n", "")
        self.chk_ts(media)
        self.chk_aac(media)
        pane = Pane(media, lines)
        self.sliding_window.slide_panes(pane)
        self.first_segment = False

    def _fixup_media(self, lines, media):
        if not media.startswith("http"):
            media = self.base_uri + "/" + media
        if self.new_media(media):
            self._parse_new_media(lines, media)

    def _post_parse(self):
        self.write_manifest()
        self.headers = []
        self.update_cue_state()
        time.sleep(self.sleep_duration)

    def _parse_manifest(self, manifest):
        """
        _parse_manifest, parses m3u8 files.
        """
        with reader(manifest) as m3u8:
            lines = []
            m3u8_lines = self.decode_lines(m3u8.readlines())
            self.mk_window_size(m3u8_lines)
            for line in m3u8_lines:
                self.chk_endlist(line)
                if line.startswith("#"):
                    if not self.parse_header(line):
                        lines.append(line)
                else:
                    media = line
                    self._fixup_media(lines, media)
                    lines = []
            self._post_parse()

    def pull(self, manifest):
        """
        pull m3u8 and parse it.
        """
        print(f"\n{SUB}{REV} Started {NORM} {iso8601()}\n")
        print(f"{SUB}{REV} Manifest {NORM} {manifest}\n")
        self.base_uri = manifest.rsplit("/", 1)[0]
        self.sliding_window = SlidingWindow()
        while self.reload:
            self._parse_manifest(manifest)
        with open(self.flat, "a") as flat:
            flat.write("#EXT-X-ENDLIST\n")


def cli():
    """
    cli is a function to use in a command line tool

        #!/usr/bin/env python3

        from fu.hls import cli

        if __name__ == "__main__":
            cli()


     is all that's required.
    """
    playlists = None
    m3u8 = None
    if "help" in sys.argv:
        print(helpme)
        sys.exit()
    if sys.argv[1].lower() == "profile":
        scp = Scte35Profile()
        scp.write_profile("hls.profile")
        sys.exit()

    with reader(sys.argv[1]) as arg:
        variants = [line for line in arg if b"#EXT-X-STREAM-INF" in line]
        if variants:
            fumo = M3uFu()
            fumo.m3u8 = sys.argv[1]
            fumo.decode()
            playlists = [
                segment
                for segment in fumo.segments
                if "#EXT-X-STREAM-INF" in segment.tags
            ]
    if playlists:
        m3u8 = playlists[0].media
    else:
        m3u8 = sys.argv[1]
    cpulr = CuePuller()
    cpulr.pull(m3u8)
    print()
    sys.exit()


helpme = """

[ fu hls ]

[ Help ]

    To display this help:

	fu hls help

[ Input ]

    fu hls takes an m3u8 URI as input.

    M3U8 formats supported:
        * master  ( When a master.m3u8 used,
                   fu hls parses the first rendition it finds )
        * rendition
    Segment types supported:
    * AAC
    * AC3
    * MPEGTS
    *codecs:
        * video
            * mpeg2, h.264, h.265
        * audio
            * mpeg2, aac, ac3, mp3

    Protocols supported:
    * file
    * http(s)
    * UDP
    * Multicast

    Encryption supported:
    * AES-128 (segments are automatically decrypted)

[ SCTE-35 ]

    fu hls displays SCTE-35 Embedded Cues as well as SCTE-35 HLS Tags.

    Supported SCTE-35:
    * All Commands, Descriptors, and UPIDS
      in the 2022-b SCTE-35 specification.

    Supported HLS Tags
    * #EXT-OATCLS-SCTE35
    * #EXT-X-CUE-OUT-CONT
    * #EXT-X-DATERANGE
    * #EXT-X-SCTE35
    * #EXT-X-CUE-IN
    * #EXT-X-CUE-OUT

[ SCTE-35 Parsing Profiles ]

    SCTE-35 parsing can be fine tuned by setting a parsing profile.

    running the command:

            fu hls profile

    will generate a default profile and write a file named hls.profile
    in the current working directory.

    a@fu:~$ cat hls.profile

    expand_cues = False
    parse_segments = False
    parse_manifests = True
    hls_tags = #EXT-OATCLS-SCTE35,#EXT-X-CUE-OUT-CONT,
    #EXT-X-DATERANGE,#EXT-X-SCTE35,#EXT-X-CUE-IN,#EXT-X-CUE-OUT
    command_types = 0x6,0x5
    descriptor_tags = 0x2
    starts = 0x22,0x30,0x32,0x34,0x36,0x44,0x46

    ( Integers are show in hex (base 16),
      base 10 unsigned integers can also be used in .35rc )

    * expand_cues:       set to True to show cues fully expanded as JSON
    * parse_segments:    set to true to enable parsing SCTE-35 from MPEGTS.
    * parse_manifests:   set to true to parse the m3u8 file for SCTE-35 HLS Tags.
    * hls_tags:          set which SCTE-35 HLS Tags to parse.
    * command_types:     set which Splice Commands to parse.
    * descriptor_tags:   set which Splice Descriptor Tags to parse.
    * starts:            set which Segmentation Type IDs to use to start breaks.

    Edit the file as needed and then run fu hls.

[ Profile Formatting Rules ]

    * Values do not need to be quoted.
    * Multiple values are separated by a commas.
    * No partial line comments. Comments must be on a separate lines.
    * Comments can be started with a # or //
    * Integers can be base 10 or base 16

[ Output Files ]

    * Created in the current working directory
    * Clobbered on start of showcues
    * Profile rules applied to the output:
        * hls.m3u8  - live playable rewrite of the m3u8 with the profile SCTE-35 rules.
        * hls.sidecar - list of ( pts, HLS SCTE-35 tag ) pairs
    * Profile rules not applied to the output:
        * hlsflat.m3u8  - hls live streams are flattened out into a vod playlist.
                     When the live m3u8  first loads, every line is written to hlsflat.m3u8
                     Wnen a live m3u8 is reloaded, everything except the headers
                    is appended to hlsflat.m3u8. This give you a VOD style m3u8
                    so you can fast forward or rewind while playing.
                    GREAT for debugging SCTE-35 live hls.

[ Cool Features ]

    * fu hls can resume when started in the middle of an ad break.

            2023-10-13T05:59:50.24Z Resuming Ad Break
            2023-10-13T05:59:50.34Z Setting Break Timer to 17.733
            2023-10-13T05:59:50.44Z Setting Break Duration to 60.067

    * mpegts streams are listed on start ( like ffprobe )

        Program: 1
            Service:
            Provider:
            Pid:	480
            Pcr Pid:	481
            Streams:
                Pid: 481[0x1e1]	Type: 0x1b AVC Video
                Pid: 482[0x1e2]	Type: 0xf AAC Audio
                Pid: 483[0x1e3]	Type: 0x86 SCTE35 Data
                Pid: 484[0x1e4]	Type: 252 Unknown
                Pid: 485[0x1e5]	Type: 0x15 ID3 Timed Meta Data

[ Example Usage ]

	* Show this help:   fu hls help

	* Generate a new hls.profile:   fu hls profile

	* parse an m3u8:    fu hls  https://example.com/out/master.m3u8

"""


if __name__ == "__main__":
    cli()



"""
sxp.py
home of the SuperXmlParser class

"""

from xml.sax.saxutils import escape, unescape
from .xml import t2s, un_camel, un_xml, strip_ns, iter_attrs
from .descriptors import descriptor_map, k_by_v
from .segmentation import table20, table22
from .upids import upid_map


class SuperXmlParser:

    def _split_attrs(self, node):
        node = node.replace("='", '="').replace("' ", '" ')
        attrs = [x for x in node.split(" ") if "=" in x]
        return attrs

    def mk_attrs(self, node):
        """
        mk_attrs parses the current node for attributes
        and stores them in self.stuff[self.active]
        """
        if "!--" in node:
            return False
        attrs = self._split_attrs(node)
        parsed = {
            x.split('="')[0]: unescape(x.split('="')[1].split('"')[0]) for x in attrs
        }
        it = iter_attrs(parsed)
        return it

    def parsed(self, tag, sub_data):
        """
        parsed returns tag, attrs and value
        for an xml element
        """
        value = ""
        attrs = {}
        if tag[0] != "/":
            tag = strip_ns(tag)
        o_data = sub_data
        if ">" in sub_data:
            ridx = sub_data.index(">")
            this_node = sub_data[: ridx + 1]
            attrs = self.mk_attrs(this_node)
            sub_data = sub_data[ridx + 1 :]
            if "<" in sub_data:
                lidx = sub_data.index("<")
                value = sub_data[:lidx].strip()
                sub_data = sub_data[lidx:].replace("\t", "")
            return {
                "tag": tag,
                "attrs": attrs,
                "this": value,
                "xml": o_data,
            }

    def mk_tag(self, data):
        """
        mk_tag parse out the
        next available xml tag from data
        """
        tag = data[1:].split(" ", 1)[0].split(">", 1)[0]
        return tag.strip()

    def mk_element(self, data, tag):
        """
        mk_element slices out the tag element
        inclusive: <tag > stuff</tag> and <tag attr="1"/>
        """
        try:
            return data[: data.index(f"</{tag}>") + len(tag) + 2]
        except:
            try:
                return data[: data.index("/>") + 1]
            except:
                print("PASS")

    def mk(self, exemel, target="SpliceInfoSection"):
        """
        parse parses an xml string for a SCTE-35 Cue.
        """
        results = []
        sub_data = None
        data = exemel.replace("\n", "").strip()
        while data:
            sub_data = None
            tag = self.mk_tag(data)
            if target in tag:  # don't even consider namespace
                sub_data = self.mk_element(data, tag)
                results.append(sub_data + ">")
                data = data.replace(sub_data, "")
            else:
                data = data.split(">", 1)[1]
                if "<" in data:
                    data = data[data.index("<") :]
                else:
                    return results

    def gimme(self, tags, exemel):
        """
        gimme get all instances of tag from exemel and return a list of them.
        """

        the_list = []
        for tag in tags:
            results = self.mk(exemel, tag)
            if results:
                element_list = [self.parsed(tag, sd) for sd in results]
                if element_list:
                    the_list.extend(element_list)
        return the_list

    def gimme_one(self, tag, exemel):
        """
        gimme_one get the first instance of tag from exemel and return it.
        """
        one = self.gimme([tag], exemel)
        if one:
            return one[0]

    def spliceinfosection(self, exemel):
        """
        spliceinfosection parses exemel for info section data
        and returns a loadable dict
        """
        my_name = "SpliceInfoSection"

        out = self.gimme_one(my_name, exemel)
        if out:
            return out["attrs"]
        return {}

    def command(self, exemel):
        """
        command parses exemel for a splice command
        and returns a loadable dict

        """
        ts = self.timesignal(exemel)
        if ts:
            return ts
        si = self.spliceinsert(exemel)
        if si:
            return si
        return {}

    def timesignal(self, exemel):
        """
        timesignal parses exemel for TimeSignal data
        and creates a loadable dict for the Cue class.
        """
        my_name = "TimeSignal"
        ts = self.gimme_one("TimeSignal", exemel)
        if ts:
            splice_time = self.splicetime(ts["xml"])
            ts["attrs"].update(splice_time)
            setme = {
                "name": "Time Signal",
                "command_type": 6,
            }
            ts["attrs"].update(setme)
            return ts["attrs"]
        return {}

    def spliceinsert(self, exemel):
        """
        spliceinsert parses exemel for SpliceInsert data
        and creates a loadable dict for the Cue class.
        """
        my_name = "SpliceInsert"
        si = self.gimme_one(my_name, exemel)
        if si:
            setme = {
                "name": my_name,
                "command_type": 5,
                "event_id_compliance_flag": True,
                "program_splice_flag": False,
                "duration_flag": False,
            }
            splice_time = self.splicetime(exemel)
            if splice_time:
                setme["program_splice_flag"] = True
            si["attrs"].update(setme)
            si["attrs"].update(splice_time)
            break_duration = self.breakduration(exemel)
            si["attrs"].update(break_duration)
            si["attrs"]["avails_expected"] = bool(si["attrs"]["avails_expected"])
            return si["attrs"]
        return {}

    def splicetime(self, exemel):
        """
        splicetime parses xml from a splice command
        to get pts_time, sets time_specified_flag to True
        """
        splicetime = self.gimme_one("SpliceTime", exemel)
        if splicetime:
            return {
                "pts_time": splicetime["attrs"]["pts_time"],
                "time_specified_flag": True,
            }
        return {}

    def breakduration(self, exemel):
        """
        breakduration parses xml for break duration, break_auto_return
        and sets duration_flag to True.
        """
        break_duration = self.gimme_one("BreakDuration", exemel)
        if break_duration:
            return {
                "break_duration": break_duration["attrs"]["duration"],
                "break_auto_return": break_duration["attrs"]["auto_return"],
                "duration_flag": True,
            }
        return {}

    def segmentationdescriptor(self, dscptr):
        """
        segmentationdescriptor creates a dict to be loaded.
        """
        my_name = "SegmentationDescriptor"
        setme = {
            "tag": 2,
            "identifier": "CUEI",
            "name": "Segmentation Descriptor",
            "segmentation_event_id_compliance_indicator": True,
            "program_segmentation_flag": True,
            "segmentation_duration_flag": False,
            "delivery_not_restricted_flag": True,
            "segmentation_event_id": hex(dscptr["attrs"]["segmentation_event_id"]),
        }
        if dscptr["attrs"]["segmentation_type_id"] in table22:
            dscptr["attrs"]["segmentation_message"] = table22[
                dscptr["attrs"]["segmentation_type_id"]
            ]
        dscptr["attrs"].update(setme)
        dr = self.deliveryrestrictions(dscptr["xml"])
        dscptr["attrs"].update(dr)
        the_upid = self.upids(dscptr["xml"])
        dscptr["attrs"].update(the_upid)
        return dscptr["attrs"]

    def upids(self, exemel):
        """
        upids parses out upids from a splice descriptors xml
        """
        ulist = self.gimme(["SegmentationUpid"], exemel)
        if len(ulist) == 1:

            seg_upid = bytes.fromhex(ulist[0]["this"].lower().replace("0x", ""))
            print(seg_upid)
            seg_upid_type = ulist[0]["attrs"]["segmentation_upid_type"]
            the_upid = {
                "segmentation_upid": ulist[0]["this"],
                "segmentation_upid_type": seg_upid_type,
                "segmentation_upid_type_name": upid_map[seg_upid_type][0],
                "segmentation_upid_length": len(seg_upid),
            }
            return the_upid

    def availdescriptor(self, dscptr):
        my_name = "AvailDescriptor"
        setme = {
            "tag": 0,
            "identifier": "CUEI",
            "name": my_name,
        }
        dscptr["attrs"].update(setme)
        return dscptr["attrs"]

    ##    def dtmfdescriptor(self,dscptr)
    ##        """
    ##        Load an DTMFDescriptor from XML
    ##        """
    ##         setme={"tag": 1,
    ##                "identifier": "CUEI",
    ##                "name":my_name,}
    ##        gonzo["DTMFDescriptor"]["dtmf_chars"] = gonzo["DTMFDescriptor"].pop("chars")
    ##        self.load(gonzo["DTMFDescriptor"])
    ##        self.dtmf_count = len(self.dtmf_chars)

    def timedescriptor(self, dscptr):
        my_name = "TimeDescriptor"
        setme = {
            "tag": 3,
            "identifier": "CUEI",
            "name": my_name,
        }
        dscptr["attrs"].update(setme)
        return dscptr["attrs"]

    def descriptors(self, exemel):
        dmap = {
            "AvailDescriptor": self.availdescriptor,
            # "DTMFDescriptor",
            "SegmentationDescriptor": self.segmentationdescriptor,
            # "TimeDescriptor",
        }
        out = []
        dlist = self.gimme(list(dmap.keys()), exemel)
        for dscptr in dlist:
            out.append(dmap[dscptr["tag"]](dscptr))
        return out

    def deliveryrestrictions(self, exemel):
        dr = self.gimme_one("DeliveryRestrictions", exemel)
        if dr:
            setme = {
                "delivery_not_restricted_flag": False,
                "device_restrictions": table20[dr["attrs"]["device_restrictions"]],
            }
            dr["attrs"].update(setme)
            return dr["attrs"]
        return {}

    def xml2cue(self, exemel):
        """
        xml2cue returns a base64 string for xmlbin
        and a dict for xml
        """
        bindata = self.gimme_one("Binary", exemel)
        if bindata:
            return bindata["this"]
        splice_info = self.spliceinfosection(exemel)
        cmd = self.command(exemel)
        dscptrs = self.descriptors(exemel)
        return {
            "info_section": splice_info,
            "command": cmd,
            "descriptors": dscptrs,
        }



"""
The bitn.BitBin and bitn.NBin classes
"""

from .spare import print2


class BitBin:
    """
    bitn.Bitbin takes a byte string and
    converts it to a integer, a very large integer
    if needed. A 1500 bit integer is no problem.
    several methods are available for slicing off bits.
    """

    def __init__(self, bites):
        # self.bites = bites
        self.bitsize = self.idx = len(bites) << 3
        self.bits = int.from_bytes(bites, byteorder="big")

    def as_90k(self, num_bits):
        """
        Returns num_bits
        of bits as 90k time
        """
        ninetyk = self.as_int(num_bits) / 90000.0
        return round(ninetyk, 6)

    def as_int(self, num_bits):
        """
        Starting at self.idx of self.bits,
        slice off num_bits of bits.
        """
        if self.idx >= num_bits:
            self.idx -= num_bits
            return (self.bits >> (self.idx)) & ~(~0 << num_bits)
        return False

    def as_hex(self, num_bits):
        """
        Returns the hex value
        of num_bits of bits
        """
        hexed = hex(self.as_int(num_bits))
        return (hexed.replace("0x", "0x0", 1), hexed)[len(hexed) % 2 == 0]

    def as_charset(self, num_bits, charset="ascii"):
        """
        Returns num_bits of bits
        as bytes decoded as charset
        default charset is ascii.
        """
        # print(charset)
        gonzo = self.as_int(num_bits)
        wide = num_bits >> 3
        if charset is None:
            return int.to_bytes(gonzo, wide, byteorder="big")
        return int.to_bytes(gonzo, wide, byteorder="big").decode(
            charset, errors="replace"
        )

    def as_bytes(self, num_bits):
        """
        Returns num_bits of bits
        as bytes
        """
        gonzo = self.as_int(num_bits)
        wide = num_bits >> 3
        return int.to_bytes(gonzo, wide, byteorder="big")

    def as_flag(self, num_bits=1):
        """
        Returns one bit as True or False
        """
        return self.as_int(num_bits) & 1 == 1

    def forward(self, num_bits):
        """
        Advances the start point
        forward by num_bits
        """
        self.idx -= num_bits

    def negative_shift(self, num_bits):
        """
        negative_shift is called instead of
        throwing a negative shift count error.
        """
        print2(f"{num_bits} bits requested, but only {self.idx} bits left.")
        print2(f"\n bytes remaining: {self.as_bytes(self.idx)} ")


class NBin:
    """
    bitn.NBin is
    the reverse BitBin.
    Encodes data to integers
    and then bytes
    """

    def __init__(self):
        self.nbits = 0
        self.idx = 0
        self.bites = b""

    def nbits2bites(self):
        """
        nbits2bites converts
        the int self.nbits to bytes as self.bites
        and sets self.nbits  and self.idx to 0
        """
        bites_wide = self.idx >> 3
        self.bites += int.to_bytes(self.nbits, bites_wide, byteorder="big")
        self.nbits = 0
        self.idx = 0

    def add_bites(self, plus_bites):
        """
        add_bites appends plus_bites
        to self.bites
        """
        if isinstance(plus_bites, int):
            plus_bites = bytes.fromhex(hex(plus_bites)[2:])
        self.bites += plus_bites

    #  if self.idx % 8 == 0:
    #     self.nbits2bites()

    def add_int(self, int_bits, bit_len):
        """
        left shift nbits and append new_bits
        """
        self.idx += bit_len
        self.nbits = (self.nbits << bit_len) | int_bits
        if self.idx % 8 == 0:
            self.nbits2bites()

    def add_90k(self, pts, bit_len=33):
        """
        Converts 90k  float timestamps
        to an int and appends it to nbits
        via self.add_int
        """
        ninetyk = int(pts * 90000.0)
        self.add_int(ninetyk, bit_len)

    def add_hex(self, hex_str, bit_len):
        """
        add_hex converts a
        hex encoded string to an int
        and appends it to self.nbits
        via self.add_int
        """
        if isinstance(hex_str, str):
            dehexed = int(hex_str, 16)
        # just in case hex_str is an int....
        else:
            dehexed = hex_str
        self.add_int(dehexed, bit_len)

    def add_flag(self, flg, bit_len=1):
        """
        add_flag takes a boolean
        value and adds it as an integer
        to self.nbits via self.add_int
        """
        bit_len = 1
        self.add_int(flg.real, bit_len)

    def reserve(self, num):
        """
        reserve sets 'num'  bits to 1
        and appends them to self.nbits
        via self.add_int
        """
        bit_len = 1
        while num:
            self.add_int(1, bit_len)
            num -= 1

    def forward(self, num):
        """
        Currently just an alias to reserve
        """
        self.reserve(num)

    def zeroed(self, num):
        """
        zeroed sets num bits to zero
        """
        bit_len = 1
        while num:
            self.add_int(0, bit_len)



#!/usr/bin/env python3

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    readme = fh.read()

with open("fu/version.py","r", encoding="utf-8") as latest:
    version = latest.read().split('"')[1]

setuptools.setup(
    name="scte35",
    version=version,
    author="Adrian of Doom",
    author_email="spam@iodisco.com",
    description="scte35 is a SCTE-35 Decoder and Encoder for MPEGTS, HLS and Everything else.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/superkabuki/scte35",
    install_requires=[
        'iframes >= 0.0.7',
        'm3ufu >= 0.0.89',
        'new_reader >= 0.1.11',
        'sideways >= 0.0.23',
        "pyaes",
    ],

    scripts=['bin/fu'],
    packages=setuptools.find_packages(),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    python_requires=">=3.6",
)
